# coding: UTF-8
"""

"""
from functools import partial

import geohash
import numpy as np
from featuretools import variable_types, primitives
from featuretools.primitives import Haversine
from sklearn.decomposition import TruncatedSVD as sk_TruncatedSVD
from sklearn.pipeline import make_pipeline

from hypernets.tabular import dask_ex as dex, sklearn_ex as skex
from hypernets.tabular.cfg import TabularCfg as cfg


class DaskCompatibleTransformPrimitive(primitives.TransformPrimitive):
    compatibility = [primitives.Library.PANDAS, primitives.Library.DASK]
    return_dtype = 'object'
    commutative = True

    def get_function(self):
        return self.fn_pd_or_dask

    def fn_pd_or_dask(self, x1, *args):
        if dex.is_dask_series(x1):
            if len(args) > 0:
                dfs = [x.to_frame() for x in [x1, *args]]
                df = dex.concat_df(dfs, axis=1)
            else:
                df = x1.to_frame()
            result = df.map_partitions(self.fn_dask_part, meta=(None, self.return_dtype))
            return result
        else:
            return self.fn_pd(x1, *args)

    def fn_pd(self, x1, *args):
        raise NotImplementedError()

    def fn_dask_part(self, part):
        arrs = [part.iloc[:, i] for i in range(len(part.columns))]
        return self.fn_pd(*arrs)


class CrossCategorical(DaskCompatibleTransformPrimitive):
    name = "cross_categorical"
    input_types = [variable_types.Categorical, variable_types.Categorical]
    return_type = variable_types.Categorical

    def fn_pd(self, x1, *args):
        result = np.array(x1, 'U')
        for x in args:
            result = np.char.add(result, np.char.add('__', np.array(x, 'U')))
        return result

    def generate_name(self, base_feature_names):
        # return "%s__%s" % (base_feature_names[0], base_feature_names[1])
        s = '__'.join(base_feature_names)
        return f'{self.name.upper()}_{s}'


class DaskCompatibleHaversine(DaskCompatibleTransformPrimitive):
    stub = Haversine(unit='kilometers')

    name = f'{stub.name}_'
    input_types = stub.input_types
    return_type = stub.return_type

    def fn_pd(self, x1, *args):
        return self.stub.get_function()(x1, args[0])

    def generate_name(self, base_feature_names):
        return self.stub.generate_name(base_feature_names)


def _geo_hash(x, precision=12):
    if any(np.isnan(x)):
        return np.nan
    else:
        return geohash.encode(x[0], x[1], precision=precision)


class GeoHashPrimitive(DaskCompatibleTransformPrimitive):
    name = "geohash"
    input_types = [variable_types.LatLong]
    return_type = variable_types.Categorical

    precision = cfg.geohash_precision

    def fn_pd(self, x1, *args):
        vfn_ = np.vectorize(partial(_geo_hash, precision=self.precision),
                            otypes=[np.object], signature='()->()')
        return vfn_(x1)


class TfidfPrimitive(primitives.TransformPrimitive):
    name = 'tfidf'
    input_types = [variable_types.NaturalLanguage]
    return_dtype = variable_types.Numeric
    commutative = True
    compatibility = [primitives.Library.PANDAS, primitives.Library.DASK]

    number_output_features = cfg.tfidf_primitive_output_feature_count
    tfidf_max_features = cfg.tfidf_max_feature_count

    def get_function(self):
        return self.fn_pd_or_dask

    def fn_pd_or_dask(self, x1):
        if isinstance(x1, dex.dd.Series):
            p = make_pipeline(dex.LocalizedTfidfVectorizer(max_features=self.tfidf_max_features),
                              dex.TruncatedSVD(n_components=self.number_output_features))
        else:
            p = make_pipeline(skex.LocalizedTfidfVectorizer(max_features=self.tfidf_max_features),
                              sk_TruncatedSVD(n_components=self.number_output_features))

        xt = p.fit_transform(x1)
        if hasattr(xt, 'iloc'):
            result = [xt.iloc[:, i] for i in range(xt.shape[1])]
        else:
            result = [xt[:, i] for i in range(xt.shape[1])]

        return result



# -*- coding:utf-8 -*-
import argparse

from hypernets.dispatchers.predict.predict_helper import PredictHelper


def main():
    parser = argparse.ArgumentParser('run predict.')
    parser.add_argument('--server', '-server',
                        default='127.0.0.1:8030',
                        help='predict server address, separated by comma')
    parser.add_argument('--chunk-size', '-chunk-size',
                        type=int, default=1000,
                        help='chunk line number')
    parser.add_argument('data_file',
                        help='data file path')
    parser.add_argument('result_file',
                        help='result file path')
    args = parser.parse_args()

    servers = list(filter(lambda s: len(s) > 0, args.server.split(',')))
    ph = PredictHelper(servers)
    ph.predict(args.data_file, args.result_file, args.chunk_size)


if __name__ == '__main__':
    try:
        main()
        print('done')
    except KeyboardInterrupt as e:
        print(e)



# -*- coding:utf-8 -*-
"""

"""
import pytest
from sklearn import preprocessing
from sklearn.compose import make_column_transformer
from sklearn.decomposition import PCA
from sklearn.impute import SimpleImputer
from sklearn.preprocessing import OneHotEncoder, PolynomialFeatures

from hypernets.tabular import sklearn_ex as skex
from hypernets.tabular.column_selector import *
from hypernets.tabular.dataframe_mapper import DataFrameMapper
from hypernets.tabular.datasets import dsutils
from hypernets.utils import const

try:
    from tensorflow.python.keras.preprocessing.sequence import pad_sequences

    tf_installed = True
except:
    tf_installed = False


def get_df():
    X = pd.DataFrame(
        {
            "a": ['a', 'b', np.nan],
            "b": list(range(1, 4)),
            "c": np.arange(3, 6).astype("u1"),
            "d": np.arange(4.0, 7.0, dtype="float64"),
            "e": [True, False, True],
            "f": pd.Categorical(['c', 'd', np.nan]),
            "g": pd.date_range("20130101", periods=3),
            "h": pd.date_range("20130101", periods=3, tz="US/Eastern"),
            "i": pd.date_range("20130101", periods=3, tz="CET"),
            "j": pd.period_range("2013-01", periods=3, freq="M"),
            "k": pd.timedelta_range("1 day", periods=3),
            "l": [1, 10, 1000]
        }
    )
    y = [1, 1, 0]
    return X, y


class Test_Transformer():
    @classmethod
    def setup_class(cls):
        cls.bank_data = dsutils.load_bank()
        cls.movie_lens = dsutils.load_movielens()

    def test_func_transformer(self):
        dfm = DataFrameMapper(
            [(column_object_category_bool, [
                SimpleImputer(strategy='constant'),
                skex.MultiLabelEncoder(),
            ]
              ),
             ],
            input_df=True,
            df_out=True,
            df_out_dtype_transforms=[
                (column_object, 'category')
            ]
        )
        X, y = get_df()
        x_new = dfm.fit_transform(X, y)
        assert x_new.dtypes.to_list() == [pd.CategoricalDtype(categories=[0, 1, 2], ordered=False),
                                          pd.CategoricalDtype(categories=[0, 1], ordered=False),
                                          pd.CategoricalDtype(categories=[0, 1, 2], ordered=False)]

    def test_pca(self):
        ct = make_column_transformer(
            (PCA(2), column_number_exclude_timedelta)
        )

        X, y = get_df()
        x_new = ct.fit_transform(X, y)
        assert x_new.shape == (3, 2)

        dfm = DataFrameMapper(
            [(column_number_exclude_timedelta, PCA(2)),
             (column_object_category_bool, [SimpleImputer(strategy='constant'), OneHotEncoder()]),
             (column_number_exclude_timedelta, PolynomialFeatures(2)),
             ], input_df=True, df_out=True
        )
        x_new = dfm.fit_transform(X, y)
        # assert x_new.columns.to_list() == ['b_c_d_l_0', 'b_c_d_l_1', 'a_a', 'a_b', 'a_missing_value', 'e_False',
        #                                    'e_True', 'f_c', 'f_d', 'f_missing_value', '1', 'b', 'c', 'd', 'l',
        #                                    'b^2', 'b c', 'b d', 'b l', 'c^2', 'c d', 'c l', 'd^2', 'd l', 'l^2']
        assert x_new.columns.to_list() == ['b_c_d_l_0', 'b_c_d_l_1', 'a_a', 'a_b', 'a_missing_value', 'e_False',
                                           'e_True', 'f_c', 'f_d', 'f_missing_value']

    def test_no_feature(self):
        df = get_df()[0]
        dfm = DataFrameMapper(
            [([], preprocessing.LabelEncoder())],
            input_df=True,
            df_out=True)

        with pytest.raises(ValueError):  # ValueError: No data output, maybe it's because your input feature is empty.
            dfm.fit_transform(df, None)

    def test_no_categorical_feature(self):
        df = get_df()[0][['b', 'd']]

        dfm = DataFrameMapper(
            [(column_object_category_bool, preprocessing.LabelEncoder())],
            input_df=True,
            df_out=True, default=None)

        x_new = dfm.fit_transform(df, None)

        assert 'b' in x_new
        assert 'd' in x_new

    def test_subsample(self):
        df = self.bank_data.copy()
        y = df.pop('y')
        X_train, X_test, y_train, y_test = skex.subsample(df, y, 100, 60, 'regression')
        assert X_train.shape == (60, 17)
        assert X_test.shape == (40, 17)
        assert y_train.shape == (60,)
        assert y_test.shape == (40,)

        X_train, X_test, y_train, y_test = skex.subsample(df, y, 100, 60, 'classification')
        assert X_train.shape == (60, 17)
        assert X_test.shape == (40, 17)
        assert y_train.shape == (60,)
        assert y_test.shape == (40,)

    def test_feature_selection(self):
        df = self.bank_data.copy()
        y = df.pop('y')
        reserved_cols = ['age', 'poutcome', 'id']
        fse = skex.FeatureSelectionTransformer('classification', 10000, 10000, 10, n_max_cols=8,
                                               reserved_cols=reserved_cols)
        fse.fit(df, y)
        assert len(fse.scores_.items()) == 10
        assert len(fse.columns_) == 11
        assert len(set(reserved_cols) - set(fse.columns_)) == 0

        x_t = fse.transform(df)
        assert x_t.columns.to_list() == fse.columns_

        df = dsutils.load_bank()
        y = df.pop('age')
        fse = skex.FeatureSelectionTransformer('regression', 10000, 10000, -1)
        fse.fit(df, y)
        assert len(fse.scores_.items()) == 17
        assert len(fse.columns_) == 10

    def test_ordinal_encoder(self):
        df1 = pd.DataFrame({"A": [1, 2, 3, 4],
                            "B": ['a', 'a', 'a', 'b']})
        df2 = pd.DataFrame({"A": [1, 2, 3, 5],
                            "B": ['a', 'b', 'z', '0']})

        ec = skex.SafeOrdinalEncoder(dtype=np.int32)
        df = ec.fit_transform(df1)
        df_expect = pd.DataFrame({"A": [0, 1, 2, 3],
                                  "B": [0, 0, 0, 1]})
        # diff = (df - df_expect).values
        # assert np.count_nonzero(diff) == 0
        assert np.where(df_expect.values == df.values, 0, 1).sum() == 0

        df = ec.transform(df2)
        df_expect = pd.DataFrame({"A": [0, 1, 2, 4],
                                  "B": [0, 1, 2, 2]})
        assert np.where(df_expect.values == df.values, 0, 1).sum() == 0

        df = ec.inverse_transform(df_expect)
        df_expect = pd.DataFrame({"A": [1, 2, 3, -1],
                                  "B": ['a', 'b', None, None]})
        assert np.where(df_expect.values == df.values, 0, 1).sum() == 0

    def test_lgbm_leaves_encoder_binary(self):
        X = self.bank_data.copy()
        y = X.pop('y')
        cats = X.select_dtypes(['int', 'int64', ]).columns.to_list()
        continous = X.select_dtypes(['float', 'float64']).columns.to_list()
        n_estimators = 50
        t = skex.LgbmLeavesEncoder(cat_vars=cats, cont_vars=continous, task=const.TASK_BINARY,
                                   n_estimators=n_estimators)
        X = t.fit_transform(X[cats + continous].copy(), y)
        assert getattr(t.lgbm, 'n_estimators', 0) > 0
        assert len(X.columns) == len(cats) + len(continous) + t.lgbm.n_estimators

    def test_lgbm_leaves_encoder_multiclass(self):
        X = self.bank_data.copy()
        y = X.pop('age')
        cats = X.select_dtypes(['int', 'int64', ]).columns.to_list()
        continous = X.select_dtypes(['float', 'float64']).columns.to_list()
        t = skex.LgbmLeavesEncoder(cat_vars=cats, cont_vars=continous, task=const.TASK_MULTICLASS)
        X = t.fit_transform(X[cats + continous].copy(), y)
        assert getattr(t.lgbm, 'n_estimators', 0) > 0
        assert len(X.columns) == len(cats) + len(continous) + t.lgbm.n_classes_ * t.lgbm.n_estimators

    def test_lgbm_leaves_encoder_regression(self):
        X = self.bank_data.copy()
        y = X.pop('age').astype('float')
        cats = X.select_dtypes(['int', 'int64', ]).columns.to_list()
        continous = X.select_dtypes(['float', 'float64']).columns.to_list()
        n_estimators = 50
        t = skex.LgbmLeavesEncoder(cat_vars=cats, cont_vars=continous, task=const.TASK_REGRESSION,
                                   n_estimators=n_estimators)
        X = t.fit_transform(X[cats + continous].copy(), y)
        assert getattr(t.lgbm, 'n_estimators', 0) > 0
        assert len(X.columns) == len(cats) + len(continous) + t.lgbm.n_estimators

    def test_cat_encoder(self):
        X = self.bank_data.copy()
        y = X.pop('y')
        cats = X.select_dtypes(['int', 'int64', ]).columns.to_list()
        t = skex.CategorizeEncoder(columns=cats)
        Xt = t.fit_transform(X.copy(), y)
        assert len(t.new_columns) == len(cats)
        assert len(Xt.columns) == len(X.columns) + len(t.new_columns)

    def test_varlen_encoder_with_movie_lens(self):
        if tf_installed:
            df = self.movie_lens.copy()
            df['genres_copy'] = df['genres']

            multi_encoder = skex.MultiVarLenFeatureEncoder([('genres', '|'), ('genres_copy', '|'), ])
            result_df = multi_encoder.fit_transform(df)

            assert multi_encoder.max_length_['genres'] > 0
            assert multi_encoder.max_length_['genres_copy'] > 0

            shape = np.array(result_df['genres'].tolist()).shape
            assert shape[1] == multi_encoder.max_length_['genres']

    def test_varlen_encoder_with_customized_data(self):
        if tf_installed:
            from io import StringIO
            data = '''
            col_foo
            a|b|c|x
            a|b
            b|b|x
            '''.replace(' ', '')

            result = pd.DataFrame({'col_foo': [
                [1, 2, 3, 4],
                [1, 2, 0, 0],
                [2, 2, 4, 0]
            ]})
            df = pd.read_csv(StringIO(data))

            multi_encoder = skex.MultiVarLenFeatureEncoder([('col_foo', '|')])
            result_df = multi_encoder.fit_transform(df)
            print(result_df)

            assert all(result_df.values == result.values)

    def test_tfidf_encoder(self):
        df = self.movie_lens.copy()
        df['genres'] = df['genres'].apply(lambda s: s.replace('|', ' '))

        encoder = skex.TfidfEncoder(['title', 'genres'], flatten=False)
        Xt = encoder.fit_transform(df.copy())
        assert 'title' in Xt.columns.tolist()
        assert 'genres' in Xt.columns.tolist()
        assert isinstance(Xt['genres'][0], (list, np.ndarray))

        encoder = skex.TfidfEncoder(flatten=True)
        Xt = encoder.fit_transform(df[['title', 'genres']].copy())
        assert 'genres' not in Xt.columns.tolist()
        assert 'genres_tfidf_0' in Xt.columns.tolist()

        encoder = skex.TfidfEncoder(flatten=False)
        Xt = encoder.fit_transform(df[['title', 'genres']].copy().values)
        assert isinstance(Xt, np.ndarray)
        assert isinstance(Xt[0, 0], (list, np.ndarray))
        assert len(Xt[0, 1]) == 19
        assert Xt.shape == (df.shape[0], 2)

        encoder = skex.TfidfEncoder(flatten=False)
        Xt = encoder.fit_transform(df[['title']].copy().values)
        assert isinstance(Xt, np.ndarray)
        assert isinstance(Xt[0, 0], (list, np.ndarray))
        assert Xt.shape == (df.shape[0], 1)

        encoder = skex.TfidfEncoder(flatten=True)
        Xt = encoder.fit_transform(df[['genres']].copy().values)
        assert isinstance(Xt, np.ndarray)
        assert isinstance(Xt[0, 0], (np.int, np.float))
        assert Xt.shape == (df.shape[0], 19)

    def test_datetime_encoder(self):
        def is_holiday(t):
            holidays = {'0501', '0502', '0503'}
            return int(t.strftime('%m%d') in holidays)

        months = {'oct': 10, 'may': 5, 'apr': 4, 'jun': 6, 'feb': 2, 'aug': 8, 'jan': 1, 'jul': 7, 'nov': 11,
                  'sep': 9, 'mar': 3, 'dec': 12}

        df = dsutils.load_bank().sample(n=1000, random_state=9527)
        df['year'] = 2000
        df['month'] = df['month'].apply(lambda s: months[s])
        df['date'] = pd.to_datetime(df[['year', 'month', 'day']])

        encoder = skex.DatetimeEncoder()
        X = encoder.fit_transform(df)
        columns = X.columns.to_list()
        assert 'date' not in columns
        assert all([c in columns for c in ['date_month', 'date_day']])
        assert all([c not in columns for c in ['date_hour', 'date_minute']])

        encoder = skex.DatetimeEncoder(include=skex.DatetimeEncoder.default_include + ['timestamp'],
                                       extra=[('holiday', is_holiday)], drop_constants=False)
        X = encoder.fit_transform(df)
        columns = X.columns.to_list()
        assert 'date' not in columns
        assert all([c in columns for c in ['date_holiday', 'date_timestamp']])



# -*- coding:utf-8 -*-

import sys
from multiprocessing import Process, Value as PValue, current_process
from os.path import getsize
from threading import Thread

from paramiko import SSHClient, AutoAddPolicy

from hypernets.utils import logging
from hypernets.utils.common import Counter

logger = logging.get_logger(__name__)


class DumpFileThread(Thread):
    counter = Counter()

    def __init__(self, in_file_handle, out_file_handle, buf_size=16):
        super(DumpFileThread, self).__init__()
        assert in_file_handle and out_file_handle

        # self.name = f'{self.__class__.__name__}-{self.counter()}'
        self.name = f'{self.__class__.__name__}-{current_process().pid}-{self.counter()}'
        self.in_file_handle = in_file_handle
        self.out_file_handle = out_file_handle
        self.buf_size = buf_size

    def run(self):
        data = self.in_file_handle.read(self.buf_size)
        while data and len(data) > 0:
            self.out_file_handle.write(data)
            data = self.in_file_handle.read(self.buf_size)


class SshProcess(Process):
    def __init__(self, ssh_host, ssh_port, cmd, in_file, out_file, err_file, environment=None):
        super(SshProcess, self).__init__()
        self.ssh_host = ssh_host
        self.ssh_port = ssh_port
        self.cmd = cmd
        self.in_file = in_file
        self.out_file = out_file
        self.err_file = err_file
        self.environment = environment
        self._exit_code = PValue('i', -1)

    def run(self):
        if logger.is_info_enabled():
            logger.info(f'[{self.name}] [SSH {self.ssh_host}]: {self.cmd}')
        try:
            code = self.ssh_run(self.ssh_host, self.ssh_port,
                                self.cmd,
                                self.in_file,
                                self.out_file,
                                self.err_file,
                                self.environment)
        except KeyboardInterrupt:
            code = 137

        if logger.is_info_enabled():
            logger.info(f'[{self.name}] [SSH {self.ssh_host}] {self.cmd} done with {code}')

        self._exit_code.value = code

    @staticmethod
    def ssh_run(ssh_host, ssh_port, cmd, in_file, out_file, err_file, environment):
        with SSHClient() as ssh:
            ssh.set_missing_host_key_policy(AutoAddPolicy())
            ssh.connect(ssh_host, ssh_port)
            stdin, stdout, stderr = ssh.exec_command(cmd, bufsize=10, environment=environment)
            if in_file and getsize(in_file) > 0:
                with open(in_file, 'rb') as f:
                    data = f.read()
                    stdin.write(data)
            stdin.flush()

            channel = stdout.channel
            # channel.settimeout(0.1)

            if out_file and err_file:
                with open(out_file, 'wb', buffering=0)as o, open(err_file, 'wb', buffering=0) as e:
                    threads = [DumpFileThread(stdout, o), DumpFileThread(stderr, e)]
                    for p in threads: p.start()
                    for p in threads: p.join()
            else:
                threads = [DumpFileThread(stdout, sys.stdout), DumpFileThread(stderr, sys.stderr)]
                for p in threads: p.start()
                for p in threads: p.join()

            assert channel.exit_status_ready()
            code = channel.recv_exit_status()
        return code

    @property
    def exitcode(self):
        code = self._exit_code.value
        return code if code >= 0 else None



# -*- coding:utf-8 -*-

import os
import re
import time
from threading import Thread

from hypernets.dispatchers.process import GrpcProcess, LocalProcess, SshProcess
from hypernets.utils import logging

logger = logging.get_logger(__name__)


def get_ip_for(peer_address):
    import socket
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        # doesn't even have to be reachable
        s.connect((peer_address, 1))
        ip = s.getsockname()[0]
    except Exception:
        ip = '127.0.0.1'
    finally:
        s.close()
    return ip


class AddressParser(object):

    def __init__(self, addr):
        super(AddressParser, self).__init__()

        pattern = r'(([a-zA-Z0-9]+)://)?([-a-zA-Z0-9+&@#/%?=~_.]+[-a-zA-Z0-9+&@#/%?=~_])(:(\d+))?'
        m = re.match(pattern, addr)
        if m is None:
            msg = f'invalid address: {addr}'
            raise Exception(msg)

        protocol = m.group(2)
        host = m.group(3)
        port = m.group(5)

        if protocol is None:
            protocol = 'ssh'
        if protocol not in {'ssh', 'grpc'}:
            raise Exception('not supported protocol: ' + protocol)

        if port is None:
            if protocol == 'ssh':
                port = '22'
            else:
                port = '8010'

        self.protocol = protocol
        self.host = host
        self.port = int(port)

    @property
    def is_ssh(self):
        return self.protocol == 'ssh'

    @property
    def is_grpc(self):
        return self.protocol == 'grpc'

    @property
    def is_localhost(self):
        return self.host == 'localhost' or self.host.startswith('127.0.')

    def __str__(self):
        return f'{self.protocol}://{self.host}:{self.port}'


class Cluster(object):

    def __init__(self, experiment,
                 driver, driver_port, start_driver,
                 executors,
                 work_dir, logs_dir,
                 report_interval,
                 *args, **kwargs):
        super(Cluster, self).__init__()
        assert isinstance(executors, (list, tuple)) and len(executors) > 0
        assert driver_port
        assert isinstance(args, (list, tuple)) and len(args) > 0

        self.experiment = experiment if experiment else time.strftime('%Y%m%d%H%M%S')
        self.driver = driver
        self.driver_port = driver_port
        self.start_driver = start_driver
        self.executors = executors
        self.work_dir = work_dir
        self.logs_dir = logs_dir
        self.report_interval = report_interval
        self.args = args
        self.kwargs = kwargs

    def run(self):
        tag = self.experiment
        os.makedirs(f'{self.logs_dir}/{tag}', exist_ok=True)

        def log_for(address):
            return f'{address.host}-{address.port}'

        def to_driver_process():
            cmd = self._driver_cmd
            a = AddressParser(self.driver if self.driver else 'localhost')
            if a.is_localhost:
                p = LocalProcess(cmd, None, None, None)
            elif a.is_grpc:
                grpc_host = f'{a.host}:{a.port}'
                p = GrpcProcess(grpc_host, cmd, None, None, None)
                return p
            else:
                p = SshProcess(a.host, a.port, cmd, None, None, None)

            p.name = 'driver'
            return p

        def to_executor_process(index):
            a = AddressParser(self.executors[index])
            cmd = self._executor_cmd

            out_file = f'{self.logs_dir}/{tag}/executor-{index}-{log_for(a)}.out'
            err_file = f'{self.logs_dir}/{tag}/executor-{index}-{log_for(a)}.err'

            if a.is_localhost:
                p = LocalProcess(cmd, None, out_file, err_file)
            elif a.is_grpc:
                grpc_host = f'{a.host}:{a.port}'
                p = GrpcProcess(grpc_host, cmd, None, out_file, err_file)
            else:
                p = SshProcess(a.host, a.port, cmd, None, out_file, err_file)

            p.name = f'executor-{index}'

            return p

        executor_processes = [to_executor_process(i) for i in range(0, len(self.executors))]
        if self.start_driver:
            driver_process = to_driver_process()
            all_processes = [driver_process, ] + executor_processes
        else:
            all_processes = executor_processes

        [p.start() for p in all_processes]

        status_thread = None
        if self.report_interval and self.report_interval > 0:
            status_thread = ClusterStatusThread(all_processes, self.report_interval)
            status_thread.start()

        [p.join() for p in all_processes]

        if status_thread:
            status_thread.stop()
            status_thread.report()

    @property
    def _driver_cmd(self):
        cmds = ' '.join(self.args)
        driver = self.driver if self.driver else '0.0.0.0'
        driver_host = AddressParser(driver).host

        ssh_cmd = f'{cmds} --driver {driver_host}:{self.driver_port} --role driver'
        if self.experiment:
            ssh_cmd += f' --experiment {self.experiment}'
        if self.work_dir:
            ssh_cmd += f' --work-dir {self.work_dir}'

        return ssh_cmd

    @property
    def _executor_cmd(self):
        cmds = ' '.join(self.args)
        if self.driver:
            a = AddressParser(self.driver)
            driver_host = a.host
        else:  # at localhost
            a = AddressParser(self.executors[0])
            driver_host = get_ip_for(a.host)

        ssh_cmd = f'{cmds} --driver {driver_host}:{self.driver_port} --role executor'
        return ssh_cmd


class ClusterStatusThread(Thread):
    def __init__(self, processes, interval=60):
        super(ClusterStatusThread, self).__init__()

        self.daemon = True

        self.processes = processes
        self.interval = interval
        self.running = False

    def run(self):
        assert not self.running

        self.running = True
        time.sleep(self.interval)

        while self.running:
            try:
                self.report()
                time.sleep(self.interval)
            except Exception as e:
                logger.error(e)

    def stop(self):
        self.running = False

    def report(self):
        def summary(p):
            if p.is_alive():
                return f'[{p.pid}] {p.name}: running'
            else:
                return f'[{p.pid}] {p.name}: done with {p.exitcode}'

        if logger.is_info_enabled():
            msg = '\n'.join([summary(p) for p in self.processes])
            logger.info('Cluster status: >>>\n' + msg + '\n<<<')



# -*- coding:utf-8 -*-
__author__ = 'yangjian'
"""

"""

from hypernets.utils.param_tuning import search_params
from hypernets.core.search_space import Choice, Real, Int
import numpy as np


def func1(p1=Choice(['a', 'b'], random_state=np.random.RandomState(9527)),
          p2=Int(1, 10, 2, random_state=np.random.RandomState(9527)),
          p3=Real(1.0, 5.0, random_state=np.random.RandomState(9527)), p4=9):
    print(f'p1:{p1},p2:{p2},p3{p3},p4:{p4}')
    return p2 * p3


class Test_ParamTuning():
    def test_search_params(self):
        print('start')
        history = search_params(func1, 'grid', max_trials=10, optimize_direction='max')
        best = history.get_best()
        assert best.reward == 14.370000000000001
        assert best.trial_no == 10



# -*- coding:utf-8 -*-
__author__ = 'yangjian'
"""

"""
import numpy as np
from deeptables.models.hyper_dt import HyperDT
from deeptables.models.hyper_dt import mini_dt_space
from hypernets.core import EarlyStoppingCallback
from hypernets.core.searcher import OptimizeDirection
from hypernets.searchers import EvolutionSearcher
from sklearn.model_selection import train_test_split

from . import BaseEstimator


class HyperDTEstimator(BaseEstimator):
    def __init__(self, task, reward_metric, max_trials=30, epochs=100, earlystop_rounds=30, time_limit=3600,
                 expected_reward=None, **kwargs):
        super(HyperDTEstimator, self).__init__(task)
        self.name = 'HyperDT'
        self.kwargs = kwargs
        self.estimator = None
        self.max_trials = max_trials
        self.reward_metric = reward_metric
        self.epochs = epochs
        self.earlystop_rounds = earlystop_rounds
        self.time_limit = time_limit
        self.expected_reward = expected_reward

    def train(self, X, y, X_test):
        searcher = EvolutionSearcher(mini_dt_space, optimize_direction=OptimizeDirection.Maximize, population_size=30,
                                     sample_size=10, regularized=True, candidates_size=10)
        es = EarlyStoppingCallback(self.earlystop_rounds, 'max', time_limit=self.time_limit,
                                   expected_reward=self.expected_reward)

        hdt = HyperDT(searcher,
                      callbacks=[es],
                      reward_metric=self.reward_metric,
                      cache_preprocessed_data=True,
                      )
        stratify = y
        if self.task == 'regression':
            stratify = None
        X_train, X_eval, y_train, y_eval = train_test_split(X, y, test_size=0.3,
                                                            random_state=9527, stratify=stratify)

        hdt.search(X_train, y_train, X_eval, y_eval, max_trials=self.max_trials, epochs=self.epochs)
        best_trial = hdt.get_best_trial()
        self.estimator = hdt.load_estimator(best_trial.model_file)

    def predict_proba(self, X):
        proba = self.estimator.predict_proba(X)
        return proba

    def predict(self, X):
        return self.estimator.predict(X)



# -*- coding:utf-8 -*-
"""

"""


class Dispatcher(object):
    def __init__(self):
        super(Dispatcher, self).__init__()

    def dispatch(self, hyper_model, X, y, X_val, y_val, cv, num_folds, max_trials, dataset_id, trial_store,
                 **fit_kwargs):
        raise NotImplemented()

    # def run_trial(self, space_sample, trial_no, X, y, X_val, y_val, **fit_kwargs):
    #     pass



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
import grpc

from hypernets.dispatchers.process.grpc.proto import proc_pb2 as hypernets_dot_dispatchers_dot_process_dot_grpc_dot_proto_dot_proc__pb2


class ProcessBrokerStub(object):
  # missing associated documentation comment in .proto file
  pass

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.run = channel.stream_stream(
        '/hypernets.dispatchers.process.grpc.proto.ProcessBroker/run',
        request_serializer=hypernets_dot_dispatchers_dot_process_dot_grpc_dot_proto_dot_proc__pb2.ProcessRequest.SerializeToString,
        response_deserializer=hypernets_dot_dispatchers_dot_process_dot_grpc_dot_proto_dot_proc__pb2.DataChunk.FromString,
        )
    self.download = channel.unary_stream(
        '/hypernets.dispatchers.process.grpc.proto.ProcessBroker/download',
        request_serializer=hypernets_dot_dispatchers_dot_process_dot_grpc_dot_proto_dot_proc__pb2.DownloadRequest.SerializeToString,
        response_deserializer=hypernets_dot_dispatchers_dot_process_dot_grpc_dot_proto_dot_proc__pb2.DataChunk.FromString,
        )


class ProcessBrokerServicer(object):
  # missing associated documentation comment in .proto file
  pass

  def run(self, request_iterator, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def download(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_ProcessBrokerServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'run': grpc.stream_stream_rpc_method_handler(
          servicer.run,
          request_deserializer=hypernets_dot_dispatchers_dot_process_dot_grpc_dot_proto_dot_proc__pb2.ProcessRequest.FromString,
          response_serializer=hypernets_dot_dispatchers_dot_process_dot_grpc_dot_proto_dot_proc__pb2.DataChunk.SerializeToString,
      ),
      'download': grpc.unary_stream_rpc_method_handler(
          servicer.download,
          request_deserializer=hypernets_dot_dispatchers_dot_process_dot_grpc_dot_proto_dot_proc__pb2.DownloadRequest.FromString,
          response_serializer=hypernets_dot_dispatchers_dot_process_dot_grpc_dot_proto_dot_proc__pb2.DataChunk.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'hypernets.dispatchers.process.grpc.proto.ProcessBroker', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))



# -*- coding:utf-8 -*-
__author__ = 'yangjian'
"""

"""

import pytest

from hypernets.core.ops import *
from hypernets.core.search_space import *
from hypernets.searchers.random_searcher import RandomSearcher


class Test_RandomSearcher():
    def get_space(self):
        space = HyperSpace()
        with space.as_default():
            p1 = Int(1, 100)
            p2 = Choice(['a', 'b'])
            p3 = Bool()
            p4 = Real(0.0, 1.0)
            id1 = Identity(p1=p1)
            id2 = Identity(p2=p2)(id1)
            id3 = Identity(p3=p3)(id2)
            id4 = Identity(p4=p4)(id3)
        return space

    def test_random_searcher(self):

        searcher = RandomSearcher(self.get_space, space_sample_validation_fn=lambda s: False)
        with pytest.raises(ValueError):
            searcher.sample()

        searcher = RandomSearcher(self.get_space, space_sample_validation_fn=lambda s: True)
        sample = searcher.sample()
        assert sample

        def valid_sample(sample):
            if sample.Param_Bool_1.value:
                return True
            else:
                return False

        searcher = RandomSearcher(self.get_space, space_sample_validation_fn=valid_sample)
        sample = searcher.sample()
        assert sample

    def test_set_random_state(self):
        from hypernets.core import set_random_state
        set_random_state(9527)

        searcher = RandomSearcher(self.get_space)
        vectors = []
        for i in range(1, 10):
            vectors.append(searcher.sample().vectors)
        assert vectors == [[98, 0, 0, 0.96], [9, 0, 0, 0.93], [60, 0, 1, 0.24], [54, 0, 1, 0.7],
                           [25, 0, 1, 0.73], [67, 1, 1, 0.43], [57, 1, 1, 0.05], [49, 0, 0, 0.71], [71, 1, 1, 0.49]]

        set_random_state(None)
        searcher = RandomSearcher(self.get_space)
        vectors = []
        for i in range(1, 10):
            vectors.append(searcher.sample().vectors)
        assert vectors != [[98, 0, 0, 0.96], [9, 0, 0, 0.93], [60, 0, 1, 0.24], [54, 0, 1, 0.7],
                           [25, 0, 1, 0.73], [67, 1, 1, 0.43], [57, 1, 1, 0.05], [49, 0, 0, 0.71], [71, 1, 1, 0.49]]

        set_random_state(9527)
        searcher = RandomSearcher(self.get_space)
        vectors = []
        for i in range(1, 10):
            vectors.append(searcher.sample().vectors)
        assert vectors == [[98, 0, 0, 0.96], [9, 0, 0, 0.93], [60, 0, 1, 0.24], [54, 0, 1, 0.7],
                           [25, 0, 1, 0.73], [67, 1, 1, 0.43], [57, 1, 1, 0.05], [49, 0, 0, 0.71], [71, 1, 1, 0.49]]

        set_random_state(1)
        searcher = RandomSearcher(self.get_space)
        vectors = []
        for i in range(1, 10):
            vectors.append(searcher.sample().vectors)
        assert vectors == [[38, 1, 0, 0.93], [10, 1, 1, 0.15], [17, 1, 0, 0.39], [7, 1, 0, 0.85], [19, 0, 1, 0.44],
                           [29, 1, 0, 0.67], [88, 1, 1, 0.43], [95, 0, 0, 0.8], [10, 1, 1, 0.09]]

        set_random_state(None)



# -*- coding:utf-8 -*-
"""

"""
from hypernets.core.search_space import *
from hypernets.core.ops import *
import pytest


def block1(step):
    id1 = Identity(name=f'{step}_block1_id_1')
    id2 = Identity(name=f'{step}_block1_id_2')(id1)
    return id2


def block2(step):
    id1_1 = Identity(name=f'{step}_block2_id_1_1')
    id1_2 = Identity(name=f'{step}_block2_id_1_2')
    id2 = Identity(name=f'{step}_block2_id_2')([id1_1, id1_2])
    return id2


def block3(step):
    id1 = Identity(name=f'{step}_block3_id_1')
    id2_1 = Identity(name=f'{step}_block2_id_2_1')(id1)
    id2_2 = Identity(name=f'{step}_block2_id_2_2')(id1)
    return [id2_1, id2_2]


ids = []


def get_id(m):
    ids.append(m.id)
    return True


class Test_ConnectionSpace:
    def test_identity(self):
        id = Identity()
        id.compile_and_forward([1,2,3])
        assert id.output, [1, 2, 3]

    def test_parameters(self):
        id = Identity()
        id.add_parameters(p1=Int(1, 10), p2=Choice([1, 2, 3]), p3=5)
        assert len(id._hyper_params) == 3
        assert id._hyper_params['p3'].value == 5

        id2 = Identity(p1=Int(1, 10), p2=Choice([1, 2, 3]), p3=5)
        assert len(id2._hyper_params) == 3
        assert id2._hyper_params['p3'].value == 5

        v1 = Int(1, 10)
        v2 = Choice([1, 2, 3])
        v3 = 5
        dict = {'p1': v1, 'p2': v2, 'p3': v3}
        id3 = Identity(**dict)
        assert len(id3._hyper_params) == 3
        assert id3._hyper_params['p3'].value == 5

        assert id3.is_params_ready == False
        assert id3.param_values == {'p1': None, 'p2': None, 'p3': 5}

        v1.assign(7)
        v2.assign(2)
        assert id3.is_params_ready == True
        assert id3.param_values == {'p1': 7, 'p2': 2, 'p3': 5}

    def test_optional(self):
        def get_space(keep_link=True):
            space = HyperSpace()
            with space.as_default():
                id1 = Identity()
                hp_opt = Bool()
                opt = Optional(Identity(), keep_link=keep_link, hp_opt=hp_opt)
                id2 = Identity()
                opt(id1)
                id2(opt)
            return space

        global ids
        space = get_space()
        ids = []
        space.traverse(get_id)
        assert ids == ['Module_Identity_1', 'Module_Optional_1', 'Module_Identity_3']

        space.Param_Bool_1.assign(True)
        ids = []
        space.traverse(get_id)
        assert ids == ['Module_Identity_1', 'Module_Identity_2', 'Module_Identity_3']

        space = get_space(keep_link=True)
        space.Param_Bool_1.assign(False)
        ids = []
        space.traverse(get_id)
        assert ids == ['Module_Identity_1', 'Module_Identity_3']

        space = get_space(keep_link=False)
        space.Param_Bool_1.assign(False)
        ids = []
        with pytest.raises(ValueError) as excinfo:
            space.traverse(get_id)
        assert excinfo.value.args[0] == 'Graph is not connected.'

    def test_optional_subgraph(self):
        def get_space(keep_link=True):
            space = HyperSpace()
            with space.as_default():
                input = Identity(name='input')
                hp_opt = Bool()
                opt = Optional(block1(1), keep_link=keep_link, hp_opt=hp_opt)(input)
                output = Identity(name='output')(opt)
                space.set_inputs(input)
            return space

        space = get_space()
        global ids
        space.Param_Bool_1.assign(True)
        ids = []
        space.traverse(get_id)
        assert ids == ['ID_input', 'ID_1_block1_id_1', 'ID_1_block1_id_2', 'ID_output']

        space = get_space()
        space.Param_Bool_1.assign(False)
        ids = []
        space.traverse(get_id)
        assert ids == ['ID_input', 'ID_output']

    def test_sequential(self):
        def get_space(keep_link=True):
            space = HyperSpace()
            with space.as_default():
                in1 = HyperInput()
                id1 = Identity()
                id2 = Identity()
                seq = Sequential([id1, id2])(in1)
                id3 = Identity()(seq)
                # id3(seq(in1))
            return space

        global ids
        space = get_space()
        ids = []
        space.traverse(get_id)
        assert ids == ['Module_HyperInput_1', 'Module_Sequential_1', 'Module_Identity_3']

        ids = []
        space.random_sample()
        space.traverse(get_id)
        assert ids == ['Module_HyperInput_1', 'Module_Identity_1', 'Module_Identity_2', 'Module_Identity_3']

    def test_sequential_subgraph(self):
        global ids
        space = HyperSpace()
        with space.as_default():
            seq = Sequential([block1(1), block1(2)])
            space.random_sample()
            ids = []
            space.traverse(get_id)
            assert ids == ['ID_1_block1_id_1', 'ID_1_block1_id_2', 'ID_2_block1_id_1', 'ID_2_block1_id_2']

        space = HyperSpace()
        with space.as_default():
            input = Identity(name='input')
            seq = Sequential([block1(1), block1(2)])(input)
            output = Identity(name='output')(seq)
            space.random_sample()
            ids = []
            space.traverse(get_id)
            assert ids == ['ID_input', 'ID_1_block1_id_1', 'ID_1_block1_id_2', 'ID_2_block1_id_1', 'ID_2_block1_id_2',
                           'ID_output']

        space = HyperSpace()
        with space.as_default():
            input = Identity(name='input')
            Sequential([block1(1), block2(1)])(input)
            ids = []
            space.random_sample()
            space.traverse(get_id)
            assert ids == ['ID_input', 'ID_1_block1_id_1', 'ID_1_block1_id_2', 'ID_1_block2_id_1_1',
                           'ID_1_block2_id_1_2',
                           'ID_1_block2_id_2']

        space = HyperSpace()
        with space.as_default():
            Sequential([block1(1), block2(1), block3(1)])
            ids = []
            space.random_sample()
            space.traverse(get_id)
            assert ids == ['ID_1_block1_id_1', 'ID_1_block1_id_2', 'ID_1_block2_id_1_1', 'ID_1_block2_id_1_2',
                           'ID_1_block2_id_2',
                           'ID_1_block3_id_1', 'ID_1_block2_id_2_1', 'ID_1_block2_id_2_2']

    def test_permutation(self):
        def get_space(keep_link=True):
            space = HyperSpace()
            with space.as_default():
                in1 = HyperInput()
                id1 = Identity()
                id2 = Identity()
                seq = Permutation([id1, id2])(in1)
                id3 = Identity()(seq)
                # id3(seq(in1))

            return space

        global ids

        space = get_space()
        ids = []

        space.traverse(get_id)
        assert ids == ['Module_HyperInput_1', 'Module_Permutation_1', 'Module_Identity_3']

        assert space.Param_Choice_1.options == [(0, 1), (1, 0)]
        ids = []
        space.Param_Choice_1.assign((1, 0))
        space.traverse(get_id)
        assert ids == ['Module_HyperInput_1', 'Module_Identity_2', 'Module_Identity_1', 'Module_Identity_3']

    def test_or(self):
        def get_space(keep_link=True):
            space = HyperSpace()
            with space.as_default():
                in1 = HyperInput()
                id1 = Identity()
                id2 = Identity()
                orop = ModuleChoice([id1, id2])
                id3 = Identity()
                orop(in1)
                id3(orop)

            return space

        global ids
        space = get_space()
        ids = []

        space.traverse(get_id)
        assert ids == ['Module_HyperInput_1', 'Module_ModuleChoice_1', 'Module_Identity_3']

        space.Module_ModuleChoice_1.hp_or.assign(0)
        ids = []
        space.traverse(get_id)
        assert ids == ['Module_HyperInput_1', 'Module_Identity_1', 'Module_Identity_3']

        space = get_space()
        space.Module_ModuleChoice_1.hp_or.assign(1)
        ids = []
        space.traverse(get_id)
        assert ids == ['Module_HyperInput_1', 'Module_Identity_2', 'Module_Identity_3']

    def test_or_subgraph(self):
        def get_space(keep_link=True):
            space = HyperSpace()
            with space.as_default():
                in1 = HyperInput(name='input')
                orop = ModuleChoice([block1(1), block2(1)])(in1)
                output = Identity(name='output')(orop)
                space.set_inputs(in1)
            return space

        global ids
        space = get_space()
        ids = []

        space.traverse(get_id)
        assert ids == ['ID_input', 'Module_ModuleChoice_1', 'ID_output']

        space.Module_ModuleChoice_1.hp_or.assign(0)
        ids = []
        space.traverse(get_id)
        assert ids == ['ID_input', 'ID_1_block1_id_1', 'ID_1_block1_id_2', 'ID_output']

        space = get_space()
        space.Module_ModuleChoice_1.hp_or.assign(1)
        ids = []
        space.traverse(get_id)
        assert ids == ['ID_input', 'ID_1_block2_id_1_1', 'ID_1_block2_id_1_2', 'ID_1_block2_id_2', 'ID_output']

    def test_repeat_id(self):
        def get_space():
            space = HyperSpace()
            with space.as_default():
                in1 = HyperInput()
                rep = Repeat(module_fn=lambda step: Identity(p1=Choice([1, 2, 3])), repeat_times=[2, 3, 4])(in1)
                id3 = Identity()(rep)
            return space

        global ids
        space = get_space()
        ids = []

        space.traverse(get_id)
        assert ids == ['Module_HyperInput_1', 'Module_Repeat_1', 'Module_Identity_1']

        assert space.Param_Choice_1.options == [2, 3, 4]
        ids = []
        space.Param_Choice_1.assign(3)
        space.traverse(get_id)
        assert ids == ['Module_HyperInput_1', 'Module_Identity_2', 'Module_Identity_3', 'Module_Identity_4',
                       'Module_Identity_1']
        assert len(space.hyper_params) == 4

    def test_repeat_seq(self):
        def seq_fn(step):
            id1 = Identity(p1=Choice([1, 2]))
            id2 = Identity(p2=Int(1, 10))
            seq = Sequential([id1, id2])
            return seq

        def get_space():
            space = HyperSpace()
            with space.as_default():
                in1 = HyperInput()
                rep = Repeat(module_fn=seq_fn, repeat_times=[2, 3, 4])(in1)
                id3 = Identity()(rep)
            return space

        space = get_space()
        global ids
        ids = []
        space.traverse(get_id)
        assert ids == ['Module_HyperInput_1', 'Module_Repeat_1', 'Module_Identity_1']

        assert space.Param_Choice_1.options == [2, 3, 4]
        ids = []
        space.Param_Choice_1.assign(3)
        space.traverse(get_id)
        assert ids == ['Module_HyperInput_1', 'Module_Sequential_1', 'Module_Sequential_2', 'Module_Sequential_3',
                       'Module_Identity_1']
        assert len(space.hyper_params) == 10

        space.random_sample()
        ids = []
        space.traverse(get_id)
        assert ids == ['Module_HyperInput_1', 'Module_Identity_2', 'Module_Identity_3', 'Module_Identity_4',
                       'Module_Identity_5', 'Module_Identity_6', 'Module_Identity_7', 'Module_Identity_1']

    def test_repeat_subgraph(self):
        def get_space():
            space = HyperSpace()
            with space.as_default():
                in1 = HyperInput()
                rep = Repeat(module_fn=block1, repeat_times=[2, 3, 4])(in1)
                id3 = Identity()(rep)
            return space

        space = get_space()
        global ids
        ids = []
        space.traverse(get_id)
        assert ids == ['Module_HyperInput_1', 'Module_Repeat_1', 'Module_Identity_1']

        assert space.Param_Choice_1.options == [2, 3, 4]
        ids = []
        space.Param_Choice_1.assign(3)
        space.traverse(get_id)
        assert ids == ['Module_HyperInput_1', 'ID_0_block1_id_1', 'ID_0_block1_id_2', 'ID_1_block1_id_1',
                       'ID_1_block1_id_2', 'ID_2_block1_id_1', 'ID_2_block1_id_2', 'Module_Identity_1']

    def test_input_choice(self):
        def get_space(keep_link=True):
            space = HyperSpace()
            with space.as_default():
                in1 = HyperInput()
                id1 = Identity()(in1)
                id2 = Identity()(in1)
                id3 = Identity()(in1)
                ic1 = InputChoice([id1, id2, id3], 2)([id1, id2, id3])
                id4 = Identity()(ic1)
            return space

        global ids
        space = get_space()
        ids = []

        space.traverse(get_id)
        assert ids == ['Module_HyperInput_1', 'Module_Identity_1', 'Module_Identity_2', 'Module_Identity_3',
                       'Module_InputChoice_1', 'Module_Identity_4']

        space.Module_InputChoice_1.hp_choice.assign([0, 1])
        assert len(space.edges - set([(space.Module_HyperInput_1, space.Module_Identity_2),
                                      (space.Module_HyperInput_1, space.Module_Identity_3),
                                      (space.Module_Identity_2, space.Module_Identity_4),
                                      (space.Module_Identity_1, space.Module_Identity_4),
                                      (space.Module_HyperInput_1, space.Module_Identity_1)])) == 0

        space = get_space()
        space.Module_InputChoice_1.hp_choice.assign([1])
        assert len(space.edges - set([(space.Module_HyperInput_1, space.Module_Identity_2),
                                      (space.Module_HyperInput_1, space.Module_Identity_3),
                                      (space.Module_Identity_2, space.Module_Identity_4),
                                      (space.Module_HyperInput_1, space.Module_Identity_1)])) == 0

        space = get_space()
        space.Module_InputChoice_1.hp_choice.assign([2])
        assert len(space.edges - set([(space.Module_HyperInput_1, space.Module_Identity_2),
                                      (space.Module_HyperInput_1, space.Module_Identity_3),
                                      (space.Module_Identity_3, space.Module_Identity_4),
                                      (space.Module_HyperInput_1, space.Module_Identity_1)])) == 0



# -*- coding:utf-8 -*-
"""

"""


class Stateful:
    def __init__(self):
        pass

    def load_state(self):
        pass

    def save_state(self):
        pass

    def get_state_path(self):
        pass



#
import copy
import pickle
from functools import partial

import numpy as np
import pandas as pd
from sklearn.linear_model import LogisticRegression
from sklearn.model_selection import KFold, StratifiedKFold
from sklearn.neural_network import MLPClassifier
from sklearn.tree import DecisionTreeClassifier

from hypernets.core import set_random_state, get_random_state
from hypernets.core.callbacks import SummaryCallback
from hypernets.core.ops import ModuleChoice, HyperInput, ModuleSpace
from hypernets.core.search_space import HyperSpace, Choice, Int, Real, Cascade, Constant, HyperNode
from hypernets.model import Estimator, HyperModel
from hypernets.searchers import make_searcher
from hypernets.tabular.dask_ex import fix_binary_predict_proba_result
from hypernets.tabular.metrics import calc_score
from hypernets.utils import fs, logging, const, infer_task_type

logger = logging.get_logger(__name__)


class PlainSearchSpace(object):
    def __init__(self, enable_dt=True, enable_lr=True, enable_nn=True):
        assert enable_dt or enable_lr or enable_nn

        super(PlainSearchSpace, self).__init__()

        self.enable_dt = enable_dt
        self.enable_lr = enable_lr
        self.enable_nn = enable_nn

    # DecisionTreeClassifier
    @property
    def dt(self):
        return dict(
            cls=DecisionTreeClassifier,
            criterion=Choice(["gini", "entropy"]),
            splitter=Choice(["best", "random"]),
            max_depth=Choice([None, 3, 5, 10, 20, 50]),
            random_state=get_random_state(),
        )

    # NN
    @property
    def nn(self):
        solver = Choice(['lbfgs', 'sgd', 'adam'])
        return dict(
            cls=MLPClassifier,
            max_iter=Int(500, 5000, step=500),
            activation=Choice(['identity', 'logistic', 'tanh', 'relu']),
            solver=solver,
            learning_rate=Choice(['constant', 'invscaling', 'adaptive']),
            learning_rate_init_stub=Cascade(partial(self._cascade, self._nn_learning_rate_init, 'slvr'), slvr=solver),
            random_state=get_random_state(),
        )

    @staticmethod
    def _nn_learning_rate_init(slvr):
        if slvr in ['sgd' or 'adam']:
            return 'learning_rate_init', Choice([0.001, 0.01])
        else:
            return 'learning_rate_init', Constant(0.001)

    # LogisticRegression
    @property
    def lr(self):
        iters = [1000]
        while iters[-1] < 9000:
            iters.append(int(round(iters[-1] * 1.25, -2)))

        solver = Choice(['newton-cg', 'lbfgs', 'liblinear', 'sag', 'saga'])
        penalty = Cascade(partial(self._cascade, self._lr_penalty_fn, 'slvr'), slvr=solver)
        l1_ratio = Cascade(partial(self._cascade, self._lr_l1_ratio, 'penalty'), penalty=penalty)

        return dict(
            cls=LogisticRegression,
            max_iter=Choice(iters),
            solver=solver,
            penalty_stub=penalty,
            l1_ratio_stub=l1_ratio,
            random_state=get_random_state(),
        )

    @staticmethod
    def _lr_penalty_fn(slvr):
        if slvr == 'saga':
            return 'penalty', Choice(['l2', 'elasticnet', 'l1', 'none'])
        else:
            return 'penalty', Constant('l2')

    @staticmethod
    def _lr_l1_ratio(penalty):
        if penalty in ['elasticnet', ]:
            return 'l1_ratio', Real(0.0, 1.0, step=0.1)
        else:
            return 'l1_ratio', Constant(None)

    # commons
    @staticmethod
    def _cascade(fn, key, args, space):
        with space.as_default():
            kvalue = args[key]
            if isinstance(kvalue, HyperNode):
                kvalue = kvalue.value
            return fn(kvalue)

    # HyperSpace
    def __call__(self, *args, **kwargs):
        space = HyperSpace()

        with space.as_default():
            hyper_input = HyperInput(name='input1')

            estimators = []
            if self.enable_dt:
                estimators.append(self.dt)
            if self.enable_lr:
                estimators.append(self.lr)
            if self.enable_nn:
                estimators.append(self.nn)

            modules = [ModuleSpace(name=f'{e["cls"].__name__}', **e) for e in estimators]
            outputs = ModuleChoice(modules)(hyper_input)
            space.set_inputs(hyper_input)

        return space


class PlainEstimator(Estimator):
    def __init__(self, space_sample, task=const.TASK_BINARY, transformer=None):
        assert task in {const.TASK_BINARY, const.TASK_MULTICLASS, const.TASK_REGRESSION}

        super(PlainEstimator, self).__init__(space_sample, task)

        # space, _ = space_sample.compile_and_forward()
        out = space_sample.get_outputs()[0]
        kwargs = out.param_values
        kwargs = {k: v for k, v in kwargs.items() if not isinstance(v, HyperNode)}

        cls = kwargs.pop('cls')
        logger.info(f'create estimator {cls.__name__}: {kwargs}')
        self.model = cls(**kwargs)
        self.cls = cls
        self.model_args = kwargs
        self.transformer = transformer

        # fitted
        self.classes_ = None
        self.cv_models_ = []

    def summary(self):
        pass

    def fit(self, X, y, **kwargs):
        eval_set = kwargs.pop('eval_set', None)  # ignore

        if self.transformer is not None:
            X = self.transformer.fit_transform(X, y)

        self.model.fit(X, y, **kwargs)
        self.classes_ = getattr(self.model, 'classes_', None)
        self.cv_models_ = []

        return self

    def fit_cross_validation(self, X, y, stratified=True, num_folds=3, shuffle=False, random_state=9527, metrics=None,
                             **kwargs):
        assert num_folds > 0
        assert isinstance(metrics, (list, tuple))

        eval_set = kwargs.pop('eval_set', None)  # ignore

        if self.transformer is not None:
            X = self.transformer.fit_transform(X, y)

        if stratified and self.task == const.TASK_BINARY:
            iterators = StratifiedKFold(n_splits=num_folds, shuffle=True, random_state=random_state)
        else:
            iterators = KFold(n_splits=num_folds, shuffle=True, random_state=random_state)

        if isinstance(y, (pd.Series, pd.DataFrame)):
            y = y.values

        oof_ = None
        oof_scores = []
        cv_models = []
        for n_fold, (train_idx, valid_idx) in enumerate(iterators.split(X, y)):
            x_train_fold, y_train_fold = X.iloc[train_idx], y[train_idx]
            x_val_fold, y_val_fold = X.iloc[valid_idx], y[valid_idx]

            fold_model = copy.deepcopy(self.model)
            fold_model.fit(x_train_fold, y_train_fold, **kwargs)

            # calc fold oof and score
            if self.task == const.TASK_REGRESSION:
                proba = fold_model.predict(x_val_fold)
                preds = proba
            else:
                proba = fold_model.predict_proba(x_val_fold)
                if self.task == const.TASK_BINARY:
                    proba = fix_binary_predict_proba_result(proba)

                proba_threshold = 0.5
                if proba.shape[-1] > 2:  # multiclass
                    preds = proba.argmax(axis=-1)
                else:  # binary:
                    preds = (proba[:, 1] > proba_threshold).astype('int32')
                preds = np.array(fold_model.classes_).take(preds, axis=0)

            if oof_ is None:
                if len(proba.shape) == 1:
                    oof_ = np.full(y.shape, np.nan, proba.dtype)
                else:
                    oof_ = np.full((y.shape[0], proba.shape[-1]), np.nan, proba.dtype)
            fold_scores = calc_score(y_val_fold, preds, proba, metrics, task=self.task)

            # save fold result
            oof_[valid_idx] = proba
            oof_scores.append(fold_scores)
            cv_models.append(fold_model)

        self.classes_ = getattr(cv_models[0], 'classes_', None)
        self.cv_models_ = cv_models

        # calc final score with mean
        scores = pd.concat([pd.Series(s) for s in oof_scores], axis=1).mean(axis=1).to_dict()
        logger.info(f'fit_cross_validation score:{scores}, folds score:{oof_scores}')

        # return
        return scores, oof_, oof_scores

    def predict(self, X, **kwargs):
        eval_set = kwargs.pop('eval_set', None)  # ignore

        if self.transformer is not None:
            X = self.transformer.transform(X)

        if self.cv_models_:
            if self.task == const.TASK_REGRESSION:
                pred_sum = None
                for est in self.cv_models_:
                    pred = est.predict(X, **kwargs)
                    if pred_sum is None:
                        pred_sum = pred
                    else:
                        pred_sum += pred
                preds = pred_sum / len(self.cv_models_)
            else:
                proba = self.predict_proba(X, **kwargs)
                preds = self.proba2predict(proba)
                preds = np.array(self.classes_).take(preds, axis=0)
        else:
            preds = self.model.predict(X, **kwargs)

        return preds

    def predict_proba(self, X, **kwargs):
        eval_set = kwargs.pop('eval_set', None)  # ignore

        if self.transformer is not None:
            X = self.transformer.transform(X)

        if self.cv_models_:
            proba_sum = None
            for est in self.cv_models_:
                proba = est.predict_proba(X, **kwargs)
                if self.task == const.TASK_BINARY:
                    proba = fix_binary_predict_proba_result(proba)
                if proba_sum is None:
                    proba_sum = proba
                else:
                    proba_sum += proba
            proba = proba_sum / len(self.cv_models_)
        else:
            proba = self.model.predict_proba(X, **kwargs)
            if self.task == const.TASK_BINARY:
                proba = fix_binary_predict_proba_result(proba)

        return proba

    def evaluate(self, X, y, metrics=None, **kwargs):
        if metrics is None:
            metrics = ['rmse'] if self.task == const.TASK_REGRESSION else ['accuracy']

        if self.task == const.TASK_REGRESSION:
            proba = None
            preds = self.predict(X, **kwargs)
        else:
            proba = self.predict_proba(X, **kwargs)
            preds = self.proba2predict(proba, proba_threshold=kwargs.get('proba_threshold', 0.5))

        scores = calc_score(y, preds, proba, metrics, self.task)
        return scores

    def proba2predict(self, proba, proba_threshold=0.5):
        if self.task == const.TASK_REGRESSION:
            return proba
        if proba.shape[-1] > 2:
            predict = proba.argmax(axis=-1)
        elif proba.shape[-1] == 2:
            predict = (proba[:, 1] > proba_threshold).astype('int32')
        else:
            predict = (proba > proba_threshold).astype('int32')
        return predict

    def save(self, model_file):
        with fs.open(model_file, 'wb') as f:
            pickle.dump(self, f, protocol=4)

    @staticmethod
    def load(model_file):
        with fs.open(model_file, 'rb') as f:
            return pickle.load(f)

    def get_iteration_scores(self):
        return []


class PlainModel(HyperModel):
    def __init__(self, searcher, dispatcher=None, callbacks=None, reward_metric=None, task=None,
                 discriminator=None, transformer=None):
        super(PlainModel, self).__init__(searcher, dispatcher=dispatcher, callbacks=callbacks,
                                         reward_metric=reward_metric, task=task)
        self.transformer = transformer

    def _get_estimator(self, space_sample):
        if callable(self.transformer):
            transformer = self.transformer()
        else:
            transformer = self.transformer

        return PlainEstimator(space_sample, task=self.task, transformer=transformer)

    def load_estimator(self, model_file):
        return PlainEstimator.load(model_file)


def train(X_train, y_train, X_eval, y_eval, task=None, reward_metric=None, optimize_direction='max', **kwargs):
    if task is None:
        task, _ = infer_task_type(y_train)
    if reward_metric is None:
        reward_metric = 'rmse' if task == const.TASK_REGRESSION else 'accuracy'

    search_space = PlainSearchSpace()
    searcher = make_searcher('mcts', search_space, optimize_direction=optimize_direction)
    callbacks = [SummaryCallback()]
    hm = PlainModel(searcher=searcher, task=task, reward_metric=reward_metric, callbacks=callbacks)
    hm.search(X_train, y_train, X_eval, y_eval, **kwargs)
    best = hm.get_best_trial()
    model = hm.final_train(best.space_sample, X_train, y_train)
    return hm, model


def train_heart_disease(**kwargs):
    from hypernets.tabular.datasets import dsutils
    from sklearn.model_selection import train_test_split

    X = dsutils.load_heart_disease_uci()
    y = X.pop('target')

    X_train, X_test, y_train, y_test = \
        train_test_split(X, y, test_size=0.3, random_state=get_random_state())
    X_train, X_eval, y_train, y_eval = \
        train_test_split(X_train, y_train, test_size=0.3, random_state=get_random_state())

    kwargs = {'reward_metric': 'auc', 'max_trials': 10, **kwargs}
    hm, model = train(X_train, y_train, X_eval, y_eval, const.TASK_BINARY, **kwargs)

    print('-' * 50)
    scores = model.evaluate(X_test, y_test, metrics=['auc', 'accuracy', 'f1', 'recall', 'precision'])
    print('scores:', scores)

    trials = hm.get_top_trials(10)
    models = [hm.load_estimator(t.model_file) for t in trials]

    msgs = [f'{t.trial_no},{t.reward},{m.cls.__name__} {m.model_args}' for t, m in zip(trials, models)]
    print('top trials:')
    print('\n'.join(msgs))


if __name__ == '__main__':
    set_random_state(335)
    train_heart_disease()



from collections import Counter
from math import ceil

from lightgbm import LGBMClassifier

from hypernets.tabular import sklearn_ex as skex, dask_ex as dex
from hypernets.tabular.datasets import dsutils
from hypernets.tabular.pseudo_labeling import sample_by_pseudo_labeling
from hypernets.tests.tabular.dask_transofromer_test import setup_dask


class Test_PseudoLabeling:

    def run_sample(self, X, y, with_dask=False):
        model = LGBMClassifier()

        if with_dask:
            setup_dask(None)
            X = dex.dd.from_pandas(X, npartitions=2)
            y = dex.dd.from_pandas(y, npartitions=2)
            model = dex.wrap_local_estimator(model)

        X_train, X_test, y_train, y_test = \
            dex.train_test_split(X, y, test_size=0.5, random_state=7)
        model.fit(X_train, y_train)
        proba = model.predict_proba(X_test)

        preds = model.predict(X_test)
        if dex.is_dask_object(preds):
            preds = dex.compute(preds)[0]
        c0 = Counter(preds)
        print('original samples:', c0)

        options = dict(threshold=0.8, number=10, quantile=0.8)
        for strategy in ['threshold', 'number', 'quantile', ]:
            X_pseudo, y_pseudo = \
                sample_by_pseudo_labeling(X_test.copy(), model.classes_, proba.copy(),
                                          strategy=strategy, **options)
            if dex.is_dask_object(y_pseudo):
                y_pseudo = dex.compute(y_pseudo)[0]

            # validate result data
            if len(y_pseudo) > 0:
                expected_y_pseudo = model.predict(X_pseudo)
                if dex.is_dask_object(expected_y_pseudo):
                    expected_y_pseudo = dex.compute(expected_y_pseudo)[0]
                assert (expected_y_pseudo == y_pseudo).all()

            # validate sample numbers
            c = Counter(y_pseudo)
            if strategy == 'number':
                assert all([v <= options['number'] for k, v in c.items()])
            elif strategy == 'quantile':
                if not with_dask:
                    expected_c = {k: ceil(c0[k] * (1 - options['quantile'])) for k, v in c0.items()}
                    assert c == expected_c

    def test_binary_sk(self):
        df = dsutils.load_bank().head(5000)
        X = skex.MultiLabelEncoder().fit_transform(df)
        y = X.pop('y')
        self.run_sample(X, y)

    def test_multiclass_sk(self):
        df = dsutils.load_bank().head(5000)
        X = skex.MultiLabelEncoder().fit_transform(df)
        y = X.pop('education')
        self.run_sample(X, y)

    def test_binary_dask(self):
        df = dsutils.load_bank().head(5000)
        X = skex.MultiLabelEncoder().fit_transform(df)
        y = X.pop('y')
        self.run_sample(X, y, with_dask=True)

    def test_multiclass_dask(self):
        df = dsutils.load_bank().head(5000)
        X = skex.MultiLabelEncoder().fit_transform(df)
        y = X.pop('education')
        self.run_sample(X, y, with_dask=True)


if __name__ == '__main__':
    pass
    # t = Foo_PseudoLabeling()
    # t.test_multiclass_sk()
    # t.test_binary_sk()
    # t.test_binary_dask()
    # t.test_multiclass_dask()



# -*- coding:utf-8 -*-
__author__ = 'yangjian'
"""

"""

import pytest

from hypernets.core.ops import *
from hypernets.core.search_space import *
from hypernets.searchers import PlaybackSearcher
from hypernets.core import TrialHistory, Trial
from hypernets.core import EarlyStoppingError

def get_space():
    space = HyperSpace()
    with space.as_default():
        id1 = Identity(p1=Choice(['a', 'b']), p2=Int(1, 100), p3=Real(0, 1.0))
    return space


th = TrialHistory('min')
sample = get_space()
sample.assign_by_vectors([0, 1, 0.1])
trial = Trial(sample, 1, 0.99, 100)
th.append(trial)

sample = get_space()
sample.assign_by_vectors([1, 2, 0.2])
trial = Trial(sample, 2, 0.9, 50)
th.append(trial)

sample = get_space()
sample.assign_by_vectors([0, 3, 0.3])
trial = Trial(sample, 3, 0.7, 200)
th.append(trial)


class Test_PlaybackSearcher():
    def test_playback_searcher(self):
        searcher = PlaybackSearcher(th, top_n=2)
        sample1 = searcher.sample()
        assert sample1.vectors == [0, 3, 0.3]
        sample2 = searcher.sample()
        assert sample2.vectors == [1, 2, 0.2]
        with pytest.raises(EarlyStoppingError) as ese:
            searcher.sample()
        assert ese.value.args[0] == 'no more samples.'



# -*- coding:utf-8 -*-
"""

"""

from collections import defaultdict

import dask.array as da
import numpy as np
import pandas as pd
from sklearn.metrics import get_scorer
from sklearn.metrics._scorer import _PredictScorer

from hypernets.utils import logging
from .base_ensemble import BaseEnsemble
from .. import dask_ex as dex

logger = logging.get_logger(__name__)


class DaskGreedyEnsemble(BaseEnsemble):

    def __init__(self, task, estimators, need_fit=False, n_folds=5, method='soft', random_state=9527,
                 scoring='neg_log_loss', ensemble_size=0, predict_kwargs=None):
        super().__init__(task, estimators, need_fit, n_folds, method, random_state)

        self.scorer = get_scorer(scoring)
        self.ensemble_size = ensemble_size
        self.predict_kwargs = predict_kwargs

        # fitted
        self.weights_ = None
        self.scores_ = None
        self.best_stack_ = None
        self.hits_ = None

    def __predict(self, estimator, X):
        if self.task == 'regression':
            if isinstance(self.predict_kwargs, dict):
                pred = estimator.predict(X, **self.predict_kwargs)
            else:
                pred = estimator.predict(X)
        else:
            if self.classes_ is None and hasattr(estimator, 'classes_'):
                self.classes_ = estimator.classes_
            if isinstance(self.predict_kwargs, dict):
                pred = estimator.predict_proba(X, **self.predict_kwargs)
            else:
                pred = estimator.predict_proba(X)
            if self.method == 'hard':
                pred = self.proba2predict(pred)
        return pred

    def fit(self, X, y_true, predictions=None):
        assert any(t is not None for t in [X, predictions])
        assert predictions is None or isinstance(predictions, (tuple, list))
        assert y_true is not None

        if predictions is None:
            predictions = [None for e in self.estimators]
        elif self.method == 'hard':
            predictions = [self.proba2predict(pred) if pred is not None else None
                           for pred in predictions]
        # else:
        #     predictions = list(predictions)

        predictions = [pred.compute() if dex.is_dask_object(pred) else pred
                       for pred in predictions]

        def get_prediction(j):
            if predictions[j] is not None:
                pred = predictions[j]
                if self.method == 'hard' and len(pred.shape) > 1 and pred.shape[1] > 1:
                    pred = self.proba2predict(pred)
            elif X is not None:
                pred = self.__predict(self.estimators[j], X)
                if dex.is_dask_object(pred):
                    pred = pred.compute()
                predictions[j] = pred
            else:
                raise ValueError(f'Not found predictions for estimator {j}')
            return pred

        if len(self.estimators) == 1:
            self.weights_ = [1]
            return self

        scores = []
        best_stack = []
        sum_predictions = None
        n_estimators = len(self.estimators)
        size = min(self.ensemble_size, n_estimators) if self.ensemble_size > 0 else n_estimators

        logger.info(f'start ensemble {n_estimators} estimators')

        if dex.is_dask_object(y_true):
            y_true = y_true.compute()

        for i in range(size):
            stack_scores = []
            stack_preds = []
            for j in range(n_estimators):
                pred = get_prediction(j)
                if sum_predictions is None:
                    sum_predictions = np.zeros(pred.shape, dtype=np.float64)
                mean_predictions = (sum_predictions + pred) / (len(best_stack) + 1)
                if isinstance(self.scorer, _PredictScorer):
                    if self.classes_ is not None:
                        # pred = np.array(self.classes_).take(np.argmax(mean_predictions, axis=1), axis=0)
                        mean_predictions = np.array(self.classes_).take(np.argmax(mean_predictions, axis=1), axis=0)
                        # mean_predictions=da.take(np.array(self.classes_),da.argmax(mean_predictions, axis=1), axis=0)
                    else:
                        mean_predictions = pred
                elif self.task == 'binary' and len(mean_predictions.shape) == 2 and mean_predictions.shape[1] == 2:
                    mean_predictions = mean_predictions[:, 1]
                # if dex.is_dask_object(mean_predictions):
                #     mean_predictions = mean_predictions.compute()
                score = self.scorer._score_func(y_true, mean_predictions, **self.scorer._kwargs) * self.scorer._sign
                stack_scores.append(score)
                stack_preds.append(pred)

            best = np.argmax(stack_scores)
            scores.append(stack_scores[best])
            best_stack.append(best)
            sum_predictions += stack_preds[best]

        # sum up estimator's hit count
        val_steps = len(best_stack)
        hits = defaultdict(int)
        for i in range(val_steps):
            hits[best_stack[i]] += 1

        weights = np.zeros((len(self.estimators)), dtype=np.float64)
        for i in range(len(self.estimators)):
            if hits.get(i) is not None:
                weights[i] = hits[i] / val_steps

        estimators_ = []
        self.weights_ = []
        for index in np.argwhere(weights > 0.).ravel():
            estimators_.append(self.estimators[index])
            self.weights_.append(weights[index])

        self.estimators = estimators_
        self.scores_ = scores
        self.hits_ = hits
        self.best_stack_ = best_stack

        logger.info(f'ensembled as {len(self.estimators)} estimators, weight:{self.weights_}')
        return self

    def predict(self, X):
        if len(self.estimators) == 1:
            preds = self.estimators[0].predict(X)
        else:
            preds = self.proba2predict(self.predict_proba(X))
            if self.task != 'regression' and self.classes_ is not None:
                preds = da.take(np.array(self.classes_), preds, axis=0)

        return preds

    def predict_proba(self, X):
        if len(self.estimators) == 1:
            proba = self.estimators[0].predict_proba(X)
        else:
            est_probas = [e.predict_proba(X) for e in self.estimators]
            est_probas = dex.compute(*est_probas)
            proba = None
            for est_proba, weight in zip(est_probas, self.weights_):
                est_proba = est_proba * weight
                if proba is None:
                    proba = est_proba
                else:
                    proba += est_proba

            if dex.is_dask_object(X):
                proba = dex.to_dask_type(proba)
        return proba

    def __repr__(self) -> str:
        if self.estimators is None:
            return 'no estimators'

        if self.weights_ is None:
            return 'not fitted'

        estimators = [getattr(e, "gbm_model", e) for e in self.estimators]
        return f'{type(self).__name__}(weight={self.weights_}, estimators={estimators})'

    def _repr_html_(self):
        estimators = [getattr(e, "gbm_model", e) for e in self.estimators] if self.estimators is not None else None
        df = pd.DataFrame([('weights', self.weights_),
                           ('scores', self.scores_),
                           ('best_stack', self.best_stack_),
                           ('hits', self.hits_),
                           ('ensemble_size', self.ensemble_size),
                           ('estimators', f'{estimators}'),
                           ])
        return df._repr_html_()



import numpy as np

from hypernets.utils import infer_task_type, const


def test_infer_task_type():
    y1 = np.random.randint(0, 2, size=(1000), dtype='int')
    y2 = np.random.randint(0, 2, size=(1000)).astype('str')
    y3 = np.random.randint(0, 20, size=(1000)).astype('object')
    y4 = np.random.random(size=(1000)).astype('float')
    y5 = np.array([1, 1, 2, 2, 'na'])

    task, _ = infer_task_type(y1)
    assert task == const.TASK_BINARY

    task, _ = infer_task_type(y2)
    assert task == const.TASK_BINARY

    task, _ = infer_task_type(y3)
    assert task == const.TASK_MULTICLASS

    task, _ = infer_task_type(y4)
    assert task == const.TASK_REGRESSION

    task, _ = infer_task_type(y5, excludes=['na'])
    assert task == const.TASK_BINARY



import pickle
import queue
import time
from threading import Thread

import grpc

from hypernets.dispatchers.cfg import DispatchCfg as c
from hypernets.utils import logging, fs
from .proto import spec_pb2_grpc
from .proto.spec_pb2 import SearchResponse, PingMessage

logger = logging.get_logger(__name__)


class TrialItem(object):
    def __init__(self, trial_no, space_file, space_sample, model_file):
        super(TrialItem, self).__init__()

        self.trial_no = trial_no
        self.space_file = space_file
        self.space_sample = space_sample
        self.space_id = space_sample.space_id
        self.model_file = model_file
        self.queue_at = time.time()

        self.success = False
        self.reward = float('nan')
        self.message = ''

    def __str__(self):
        return f'{self.__dict__}'


class SearchHolder(object):
    def __init__(self, search_id, spaces_dir, models_dir, on_next, on_report, on_summary):
        super(SearchHolder, self).__init__()

        fs.makedirs(spaces_dir, exist_ok=True)
        fs.makedirs(models_dir, exist_ok=True)

        self.search_id = search_id
        self.spaces_dir = spaces_dir
        self.models_dir = models_dir

        self.on_next = on_next
        self.on_report = on_report
        self.on_summary = on_summary

        self.start_at = time.time()
        self.finish_at = None

        self.queued_pool = queue.Queue()  # TrialItem
        self.running_items = {}  # space_id -> TrialItem
        self.reported_items = {}  # space_id -> TrialItem
        self.all_items = {}  # space_id -> TrialItem

    def add(self, trial_no, space_sample):
        space_id = space_sample.space_id
        assert space_id not in self.all_items.keys()

        space_file = f'{self.spaces_dir}/space-{trial_no}.pkl'
        with fs.open(space_file, 'wb') as f:
            pickle.dump(space_sample, f)

        model_file = '%s/%05d_%s.pkl' % (self.models_dir, trial_no, space_id)
        item = TrialItem(trial_no, space_file, space_sample, model_file)

        detail = f'trial_no={item.trial_no}, space_id={item.space_id}, space_file={space_file}'
        if logger.is_info_enabled():
            logger.info(f'[{self.search_id}] [search] {detail}')

        self.queued_pool.put(item)
        self.all_items[space_id] = item

    def readd(self, item):
        space_id = item.space_id
        space_file = item.space_file
        assert space_id in self.all_items.keys()

        detail = f'trial_no={item.trial_no}' \
                 + f', space_id={item.space_id}, space_file={space_file}'
        if logger.is_info_enabled():
            logger.info(f'[{self.search_id}] [re-push] {detail}')

        if space_id in self.running_items.keys():
            if logger.is_info_enabled():
                logger.info(f'[remove running] {detail}')
            self.running_items.pop(space_id)

        if space_id in self.reported_items.keys():
            if logger.is_info_enabled():
                logger.info(f'[remove reported] {detail}')
            self.reported_items.pop(space_id)

        self.queued_pool.put(item)

    def running_size(self):
        return len(self.all_items) - len(self.reported_items)

    def queue_size(self):
        return self.queued_pool.qsize()

    @property
    def running(self):
        return len(self.all_items) > len(self.reported_items)

    def get_next_item(self, peer, wait_hook):
        while self.running:
            try:
                item = self.queued_pool.get(False)
                item.peer = peer
                item.start_at = time.time()
                detail = f'trial_no={item.trial_no}, space_id={item.space_id}' \
                         + f',space_file={item.space_file}'
                if logger.is_info_enabled():
                    logger.info(f'[{self.search_id}] [dispatch] [{peer}] {detail}')
                return item
            except queue.Empty:
                time.sleep(0.1)
                if wait_hook():
                    continue
                else:
                    break

        return None

    def report_item(self, peer, space_id, success, reward, message):
        assert space_id in self.all_items.keys()

        item = self.all_items[space_id]
        detail = f'trial_no={item.trial_no}, space_id={space_id}' \
                 + f', reward={reward}, success={success}'
        if not success:
            detail += f', message={message}'

        if space_id not in self.running_items.keys():
            msg = f'[{self.search_id}] [ignored-not running-report] [{peer}] {detail}'
            logger.warning(msg)
        else:
            item.success = success
            item.reward = reward
            item.message = message
            item.report_at = time.time()

            self.running_items.pop(space_id)
            self.reported_items[space_id] = item

            if logger.is_info_enabled():
                logger.info(f'[{self.search_id}] [report] [{peer}] {detail}')

            if self.on_report:
                try:
                    self.on_report(item)
                except Exception:
                    import traceback
                    traceback.print_exc()


class SearchDriverService(spec_pb2_grpc.SearchDriverServicer):
    def __init__(self, spaces_dir, models_dir):
        super(SearchDriverService, self).__init__()

        self.spaces_dir = spaces_dir
        self.models_dir = models_dir

        self.searches = []  # SearchHolder

        self.status_thread = DriverStatusThread(self)
        self.status_thread.start()

    def __del__(self):
        self.status_thread.stop()

    def _find_search(self, search_id):
        for s in self.searches:
            if s.search_id == search_id:
                return s
        return None

    def ping(self, request, context):
        peer = context.peer()
        message = request.message
        if logger.is_info_enabled():
            logger.info(f'[ping] [{peer}] {message}')
        return PingMessage(message=message)

    def search(self, request_iterator, context):
        def response_with(search_id, item):
            return SearchResponse(code=SearchResponse.OK,
                                  search_id=search_id,
                                  trial_no=str(item.trial_no),
                                  space_id=item.space_id,
                                  space_file=item.space_file,
                                  model_file=item.model_file)

        def response_finished(search_id):
            return SearchResponse(code=SearchResponse.FINISHED,
                                  search_id=search_id,
                                  trial_no='',
                                  space_id='',
                                  space_file='',
                                  model_file='')

        def response_waiting(search_id):
            return SearchResponse(code=SearchResponse.WAITING,
                                  search_id=search_id,
                                  trial_no='',
                                  space_id='',
                                  space_file='',
                                  model_file='')

        def response_failed(search_id):
            return SearchResponse(code=SearchResponse.FAILED,
                                  search_id=search_id,
                                  trial_no='',
                                  space_id='',
                                  space_file='',
                                  model_file='')

        peer = context.peer()
        running_item = None

        try:
            for request in request_iterator:
                search_id = request.search_id
                space_id = request.space_id

                search = self.current_search
                if search_id != search.search_id:
                    assert running_item is None

                    if self._find_search(search_id):
                        msg = response_finished(search_id)
                        yield msg
                        break
                    else:
                        # not found, maybe future search, make it wait
                        msg = response_waiting(search_id)
                        yield msg
                        continue

                if running_item:
                    assert space_id == running_item.space_id
                    success = request.success
                    reward = request.reward
                    message = request.message
                    search.report_item(peer, space_id, success, reward, message)
                    running_item = None

                item = search.get_next_item(peer, wait_hook=context.is_active)
                if item:
                    assert item.space_id not in search.running_items.keys()

                    msg = response_with(search_id, item)
                    search.running_items[item.space_id] = item
                    running_item = item
                    yield msg
                else:
                    break
        except grpc.RpcError as e:
            # ignore, just log it
            import traceback
            trace_detail = traceback.format_exc()
            try:
                msg = f'RpcError {peer} {e.__class__.__name__}: {e.code()}'
                logger.warning(msg)
            except Exception:
                msg = f'RpcError {peer} {e.__class__.__name__}:\n'
                logger.error(msg + trace_detail)
        except Exception as e:
            logger.error(f'{e.__class__.__name__}: {e}')
            import traceback
            traceback.print_exc()
        finally:
            if running_item:
                self.current_search.readd(running_item)

    def start_search(self, search_id, on_next, on_report, on_summary):
        old_search = self._find_search(search_id)
        assert old_search is None

        search_spaces_dir = f'{self.spaces_dir}/{search_id}'
        search_models_dir = f'{self.models_dir}/{search_id}'
        search = SearchHolder(search_id, search_spaces_dir, search_models_dir, on_next, on_report, on_summary)
        self.searches.append(search)

        if logger.is_info_enabled():
            logger.info(f'>>>enter {search_id}')

    @property
    def current_search(self):
        assert len(self.searches) > 0
        return self.searches[-1]

    def add(self, trial_no, space_sample):
        self.current_search.add(trial_no, space_sample)

    def running_size(self):
        return self.current_search.running_size()

    def queue_size(self):
        return self.current_search.queue_size()


class DriverStatusThread(Thread):
    def __init__(self, service):
        super(DriverStatusThread, self).__init__()

        self.service = service
        self.daemon = True
        self.running = False
        self.summary_interval = c.cluster_summary_interval

    def run(self) -> None:
        self.running = True

        summary_at = time.time()
        while self.running:
            try:
                now = time.time()
                if self.report_summary and self.summary_interval > 0 and summary_at + self.summary_interval < now:
                    self.report_summary()
                    summary_at = now

                time.sleep(1)
            except Exception:
                import traceback
                traceback.print_exc()

        if logger.is_info_enabled():
            logger.info('SearchDriverService shutdown')

    def stop(self):
        self.running = False

    def report_summary(self):
        if not logger.is_info_enabled():
            return

        search = self.service.current_search
        search_id = search.search_id
        total = len(search.all_items)
        running = len(search.running_items)
        reported = len(search.reported_items)
        queued = total - running - reported

        msg = f'[{search_id}] [summary] queued={queued}, running={running}, reported={reported}, total={total}'
        if running > 0:
            detail = [f'trial-{x.trial_no}@{x.peer}' for x in search.running_items.values()]
            msg += '\n\trunning: ' + ','.join(detail)

        if search.on_summary:
            service_summary = search.on_summary()
            if service_summary:
                msg += '\n\t' + service_summary

        logger.info(msg)


_grpc_servers = {}  # add -> tuple(grpc_server, driver_service)


def serve(addr, search_id, spaces_dir, models_dir, on_next=None, on_report=None, on_summary=None):
    import grpc
    from concurrent import futures

    worker_number = c.grpc_worker_count
    service = SearchDriverService(spaces_dir, models_dir)
    server = grpc.server(futures.ThreadPoolExecutor(max_workers=worker_number))
    spec_pb2_grpc.add_SearchDriverServicer_to_server(service, server)

    server.add_insecure_port(addr)
    server.start()

    service.start_search(search_id, on_next, on_report, on_summary)

    return server, service


def get_or_serve(addr, search_id, spaces_dir, models_dir, on_next=None, on_report=None, on_summary=None):
    s = _grpc_servers.get(addr)

    if s is None:
        server, service = serve(addr, search_id, spaces_dir, models_dir, on_next, on_report, on_summary)
        _grpc_servers[addr] = (server, service)
    else:
        server, service = s
        service.start_search(search_id, on_next, on_report, on_summary)

    return server, service



# -*- coding:utf-8 -*-
__author__ = 'yangjian'
"""

"""
from hypernets.core.ops import Identity, Choice
from hypernets.core.search_space import HyperSpace, Int, Real
from hypernets.core.trial import TrialHistory, Trial
from .. import test_output_dir


class Test_TrialHistory():
    def test_is_exsited(self):
        def get_space():
            space = HyperSpace()
            with space.as_default():
                id1 = Identity(p1=Choice(['a', 'b']), p2=Int(1, 100), p3=Real(0, 1.0))
            return space

        th = TrialHistory('min')
        sample = get_space()
        sample.random_sample()
        trial = Trial(sample, 1, 0.99, 100)
        th.append(trial)

        sample2 = get_space()
        sample2.assign_by_vectors(sample.vectors)

        assert th.is_existed(sample2)

        t = th.get_trial(sample2)
        assert t.reward == 0.99
        assert t.elapsed == 100
        assert t.trial_no == 1

        sample3 = get_space()
        sample3.random_sample()

        assert not th.is_existed(sample3)

    def test_save_load(self):
        def get_space():
            space = HyperSpace()
            with space.as_default():
                id1 = Identity(p1=Choice(['a', 'b']), p2=Int(1, 100), p3=Real(0, 1.0))
            return space

        th = TrialHistory('min')
        sample = get_space()
        sample.assign_by_vectors([0, 1, 0.1])
        trial = Trial(sample, 1, 0.99, 100)
        th.append(trial)

        sample = get_space()
        sample.assign_by_vectors([1, 2, 0.2])
        trial = Trial(sample, 2, 0.9, 50)
        th.append(trial)

        sample = get_space()
        sample.assign_by_vectors([0, 3, 0.3])
        trial = Trial(sample, 3, 0.7, 200)
        th.append(trial)

        filepath = f'{test_output_dir}/history.txt'
        th.save(filepath)

        with open(filepath) as f:
            lines = f.readlines()
            # assert lines == ['min\n', '1|[0, 1, 0.1]|0.99|100\n', '2|[1, 2, 0.2]|0.9|50\n', '3|[0, 3, 0.3]|0.7|200\n']

        history = TrialHistory.load_history(get_space, filepath)
        assert history.optimize_direction == 'min'
        assert len(history.trials) == 3
        assert history.trials[0].space_sample.vectors == [0, 1, 0.1]
        assert history.trials[0].elapsed == 100.0
        assert history.trials[0].reward == 0.99
        assert history.trials[0].trial_no == 1

        trajectories = history.get_trajectories()
        assert trajectories == ([0.0, 100.0, 150.0, 350.0], [0.0, 0.99, 0.99, 0.99], [0.0, 0.99, 0.9, 0.7], 1, 100.0)



import inspect
import pickle
from functools import partial

import dask.array as da
import dask.dataframe as dd
import pandas as pd
from sklearn.base import BaseEstimator

from hypernets import __version__
from hypernets.utils import fs, hash_data, logging
from .cfg import TabularCfg as cfg
from .persistence import to_parquet, read_parquet

logger = logging.get_logger(__name__)

_STRATEGY_DATA = 'data'
_STRATEGY_TRANSFORM = 'transform'

_KIND_DEFAULT = 'pickle'
_KIND_DATAFRAME = 'dataframe'
_KIND_LIST = 'list'
_KIND_NONE = 'none'

_KIND_DASK_ARRAY = 'dask_array'
_KIND_DASK_DATAFRAME = 'dask_dataframe'
_KIND_DASK_SERIES = 'dask_series'


class CacheCallback:
    def on_enter(self, fn, *args, **kwargs):
        """
        is fired before checking cache.
        raise Exception to skip load cache
        """
        pass

    def on_apply(self, fn, cached_data, *args, **kwargs):
        """
        is fired before applying cached data.
        raise Exception to skip applying
        """
        pass

    def on_store(self, fn, cached_data, *args, **kwargs):
        """
        is fired before storing cache.
        raise Exception to skip store cache
        """
        pass

    def on_leave(self, fn, *args, **kwargs):
        """
        is fired before leaving fn call.
        raise Exception to skip store cache
        """
        pass


def cache(strategy=None, arg_keys=None, attr_keys=None, attrs_to_restore=None, transformer=None,
          callbacks=None, cache_dir=None):
    assert strategy in [_STRATEGY_TRANSFORM, _STRATEGY_TRANSFORM, None]
    assert isinstance(arg_keys, (tuple, list, str, type(None)))
    assert isinstance(attr_keys, (tuple, list, str, type(None)))
    assert isinstance(attrs_to_restore, (tuple, list, str, type(None)))
    assert callable(transformer) or isinstance(transformer, str) or transformer is None
    assert callbacks is None or isinstance(callbacks, CacheCallback) \
           or all([issubclass(type(c), CacheCallback) for c in callbacks])

    if isinstance(arg_keys, str):
        arg_keys = [a.strip(' ') for a in arg_keys.split(',') if len(a.strip(' ')) > 0]
    if isinstance(attr_keys, str):
        attr_keys = [a.strip(' ') for a in attr_keys.split(',') if len(a.strip(' ')) > 0]
    if isinstance(attrs_to_restore, str):
        attrs_to_restore = [a.strip(' ') for a in attrs_to_restore.split(',') if len(a.strip(' ')) > 0]

    if isinstance(callbacks, CacheCallback):
        callbacks = [callbacks]

    return partial(decorate,
                   strategy=strategy,
                   cache_dir=cache_dir,
                   attr_keys=attr_keys,
                   arg_keys=arg_keys,
                   attrs_to_restore=attrs_to_restore,
                   transformer=transformer,
                   callbacks=callbacks)


def decorate(fn, *, cache_dir, strategy,
             arg_keys=None, attr_keys=None, attrs_to_restore=None,
             transformer=None, callbacks=None):
    assert callable(fn)

    sig = inspect.signature(fn)
    if isinstance(transformer, str) or attr_keys is not None or attrs_to_restore is not None:
        assert 'self' in sig.parameters.keys()

    if cfg.cache_strategy == 'disabled':
        return fn

    if callbacks is None:
        callbacks = []

    if cache_dir is None:
        cache_dir = f'{cfg.cache_dir}{fs.sep}{".".join([fn.__module__, fn.__qualname__])}'

    if cfg.cache_strategy != 'disabled' and not fs.exists(cache_dir):
        try:
            fs.mkdirs(cache_dir, exist_ok=True)
        except:
            logger.warning(f'Failed to create cache directory "{cache_dir}".')

    def _cache_call(*args, **kwargs):
        assert len(args) > 0

        obj = None
        cache_path = None
        loaded = False
        result = None

        try:
            for c in callbacks:
                c.on_enter(fn, *args, **kwargs)

            # bind arguments
            bind_args = sig.bind(*args, **kwargs)
            bind_args.apply_defaults()

            obj = bind_args.arguments.get('self', None)

            # calc cache_key
            key_items = {}

            arg_kwargs = bind_args.arguments.get('kwargs', {}).copy()
            arg_items = {k: v for k, v in bind_args.arguments.items() if k not in ['self', ]}  # as dict
            arg_items.update(arg_kwargs)

            if arg_keys is not None and len(arg_keys) > 0:
                key_items.update({k: arg_items.get(k) for k in arg_keys})
            else:
                key_items.update(arg_items)

            if attr_keys is not None:
                key_items.update({k: getattr(obj, k, None) for k in attr_keys})
            elif isinstance(obj, BaseEstimator) and 'params_' not in key_items:
                key_items['params_'] = obj.get_params(deep=False)

            if attrs_to_restore is not None:
                key_items['attrs_to_restore_'] = attrs_to_restore

            cache_key = hash_data(key_items)

            # join cache_path
            if not fs.exists(cache_dir):
                fs.mkdirs(cache_dir, exist_ok=True)
            cache_path = f'{cache_dir}{fs.sep}{cache_key}'

            # detect and load cache
            if fs.exists(f'{cache_path}.meta'):
                # load
                cached_data, meta = _load_cache(cache_path)

                for c in callbacks:
                    c.on_apply(fn, cached_data, *args, **kwargs)

                # restore attributes
                if attrs_to_restore is not None:
                    cached_attributes = meta.get('attributes', {})
                    for k in attrs_to_restore:
                        setattr(obj, k, cached_attributes.get(k))

                if meta['strategy'] == 'data':
                    result = cached_data
                else:  # strategy==transform
                    if isinstance(transformer, str):
                        tfn = getattr(obj, transformer)
                        assert callable(tfn)
                        result = tfn(*args[1:], **kwargs)  # exclude args[0]==self
                    elif callable(transformer):
                        result = transformer(*args, **kwargs)

                loaded = True
        except Exception as e:
            logger.warning(e)

        if not loaded:
            result = fn(*args, **kwargs)

        if cache_path is not None and not loaded:
            try:
                for c in callbacks:
                    c.on_store(fn, result, *args, **kwargs)

                # store cache
                cache_strategy = strategy if strategy is not None else cfg.cache_strategy
                if cache_strategy == 'transform' and (result is None or transformer is not None):
                    cache_data = None
                    meta = {'strategy': 'transform'}
                else:
                    cache_data = result
                    meta = {'strategy': 'data'}

                if attrs_to_restore is not None:
                    meta['attributes'] = {k: getattr(obj, k, None) for k in attrs_to_restore}
                if isinstance(obj, BaseEstimator):
                    meta['params_'] = obj.get_params(deep=False)  # for info

                _store_cache(cache_path, cache_data, meta=meta)

                for c in callbacks:
                    c.on_leave(fn, *args, **kwargs)

            except Exception as e:
                logger.warning(e)

        return result

    return _cache_call


def _store_cache(cache_path, data, meta):
    meta = meta.copy() if meta is not None else {}
    meta['version'] = __version__

    if isinstance(data, (list, tuple)):
        items = [f'_{i}' for i in range(len(data))]
        for d, i in zip(data, items):
            _store_cache(f'{cache_path}{i}', d, meta)
        meta.update({'kind': _KIND_LIST, 'items': items})
    elif isinstance(data, pd.DataFrame):
        item = f'.parquet'
        to_parquet(data, f'{cache_path}{item}', delayed=False, filesystem=fs)
        meta.update({'kind': _KIND_DATAFRAME, 'items': [item]})
    elif data is None:
        meta.update({'kind': _KIND_NONE, 'items': []})
    elif isinstance(data, dd.DataFrame):
        item = f'.parquet'
        if not fs.exists(f'{cache_path}{item}'):
            fs.mkdirs(f'{cache_path}{item}')
        to_parquet(data, f'{cache_path}{item}', delayed=False, filesystem=fs)
        meta.update({'kind': _KIND_DASK_DATAFRAME, 'items': [item]})
    elif isinstance(data, dd.Series):
        item = f'.parquet'
        df = data.to_frame()
        if not fs.exists(f'{cache_path}{item}'):
            fs.mkdirs(f'{cache_path}{item}')
        to_parquet(df, f'{cache_path}{item}', delayed=False, filesystem=fs)
        meta.update({'kind': _KIND_DASK_SERIES, 'items': [item]})
    elif isinstance(data, da.Array):
        item = f'.parquet'
        columns = [f'c{i}' for i in range(data.shape[-1])]
        df = dd.from_dask_array(data, columns=columns)
        if not fs.exists(f'{cache_path}{item}'):
            fs.mkdirs(f'{cache_path}{item}')
        to_parquet(df, f'{cache_path}{item}', delayed=False, filesystem=fs)
        meta.update({'kind': _KIND_DASK_ARRAY, 'items': [item]})
    else:
        item = f'.pkl'
        with fs.open(f'{cache_path}{item}', 'wb') as f:
            pickle.dump(data, f, protocol=pickle.HIGHEST_PROTOCOL)
        meta.update({'kind': _KIND_DEFAULT, 'items': [item]})

    with fs.open(f'{cache_path}.meta', 'wb') as f:
        pickle.dump(meta, f, protocol=pickle.HIGHEST_PROTOCOL)


def _load_cache(cache_path):
    with fs.open(f'{cache_path}.meta', 'rb') as f:
        meta = pickle.load(f)

    if meta['version'] != __version__:
        raise EnvironmentError(f'Incompatible version: {meta["version"]}')

    data_kind = meta['kind']
    items = meta['items']

    if data_kind == _KIND_LIST:
        data = [_load_cache(f'{cache_path}{i}')[0] for i in items]
    elif data_kind == _KIND_NONE:
        data = None
    elif data_kind == _KIND_DEFAULT:  # pickle
        with fs.open(f'{cache_path}{items[0]}', 'rb') as f:
            data = pickle.load(f)
    elif data_kind == _KIND_DATAFRAME:
        data = read_parquet(f'{cache_path}{items[0]}', delayed=False, filesystem=fs)
    elif data_kind == _KIND_DASK_DATAFRAME:
        data = read_parquet(f'{cache_path}{items[0]}', delayed=True, filesystem=fs)
    elif data_kind == _KIND_DASK_SERIES:
        df = read_parquet(f'{cache_path}{items[0]}', delayed=True, filesystem=fs)
        data = df[df.columns[0]]
    elif data_kind == _KIND_DASK_ARRAY:
        df = read_parquet(f'{cache_path}{items[0]}', delayed=True, filesystem=fs)
        data = df.to_dask_array(lengths=True)
    else:
        raise ValueError(f'Unexpected cache data kind "{data_kind}"')

    return data, meta


def clear(cache_dir=None, fn=None):
    assert fn is None or callable(fn)

    if cache_dir is None:
        cache_dir = cfg.cache_dir
    if callable(fn):
        cache_dir = f'{cache_dir}{fs.sep}{".".join([fn.__module__, fn.__qualname__])}'

    if fs.exists(cache_dir):
        fs.rm(cache_dir, recursive=True)
        fs.mkdirs(cache_dir, exist_ok=True)
#
#
# class CacheLoader:
#     def __call__(self, cache_path):
#         raise NotImplemented
#
#     def accept(self, meta):
#         raise NotImplemented
#
#
# class CacheStorer:
#     def __call__(self, cache_path):
#         raise NotImplemented
#
#     def accept(self, data):
#         raise NotImplemented



# -*- coding:utf-8 -*-

import math
import queue
import time
from concurrent.futures import ThreadPoolExecutor, as_completed

import dask
from dask.distributed import Client, default_client

from hypernets.core.callbacks import EarlyStoppingError
from hypernets.core.dispatcher import Dispatcher
from hypernets.core.trial import Trial
from hypernets.dispatchers.cfg import DispatchCfg as c
from hypernets.utils import logging, fs
from hypernets.utils.common import Counter

logger = logging.get_logger(__name__)


class DaskTrialItem(Trial):
    def __init__(self, space_sample, trial_no, reward=math.nan, elapsed=math.nan, model_file=None):
        super(DaskTrialItem, self).__init__(space_sample, trial_no,
                                            reward, elapsed,
                                            model_file)

        self.space_id = space_sample.space_id
        self.queue_at = time.time()

    def __str__(self):
        return f'{self.__dict__}'

    @staticmethod
    def copy_from(other):
        return DaskTrialItem(other.space_sample,
                             other.trial_no,
                             other.reward,
                             other.elapsed,
                             other.model_file)


class DaskExecutorPool(object):
    def __init__(self, worker_count, queue_size,
                 on_trial_start, on_trial_done,
                 fn_run_trial,
                 X, y, X_val, y_val, trial_kwargs):
        self.on_trial_start = on_trial_start
        self.on_trial_done = on_trial_done

        self.fn_run_trial_delayed = dask.delayed(fn_run_trial)
        self.X_delayed = dask.delayed(X)
        self.y_delayed = dask.delayed(y)
        self.X_val_delayed = dask.delayed(X_val)
        self.y_val_delayed = dask.delayed(y_val)
        self.trial_kwargs = trial_kwargs

        self.running = False
        self.interrupted = False

        self.worker_count = worker_count
        self.queue_size = queue_size

        self.pool = ThreadPoolExecutor(max_workers=worker_count)
        self.queue = queue.Queue(queue_size)
        self.tasks = []

    @property
    def done(self):
        return all(t.done for t in self.tasks)

    @property
    def qsize(self):
        return self.queue.qsize()

    def push(self, trial_item):
        self.queue.put(trial_item)

    def start(self):
        assert len(self.tasks) == 0

        self.running = True
        self.tasks = [self.pool.submit(self._wait_and_run, self.queue, self._run_trial)
                      for _ in range(self.worker_count)]

        return self.tasks

    def stop(self):
        self.running = False

    def join(self):
        for f in as_completed(self.tasks):
            # logger.info(f'Handle trial count: {f.result()}')
            pass

    def cancel(self):
        self.running = False
        self.pool.shutdown(True)

    def _wait_and_run(self, trial_queue, fn):
        n = 0
        while self.running:
            trial_item = trial_queue.get(block=True)
            if self.running and trial_item is not None:
                fn(trial_item)
                n += 1
            else:
                # invoke others
                self.queue.put(trial_item)
                break

        return n

    def _run_trial(self, trial_item):
        try:
            if self.on_trial_start:
                self.on_trial_start(trial_item)

            fn = self.fn_run_trial_delayed
            d = fn(trial_item.space_sample,
                   trial_item.trial_no,
                   self.X_delayed, self.y_delayed,
                   self.X_val_delayed, self.y_val_delayed,
                   trial_item.model_file,
                   **self.trial_kwargs)
            result = d.compute()

            trial_item.reward = result.reward
            trial_item.elapsed = result.elapsed

            if self.on_trial_done:
                self.on_trial_done(trial_item)

        except KeyboardInterrupt:
            self.running = False
            self.interrupted = True
            self.queue.put(None)  # mark end
            print('KeyboardInterrupt')
        except Exception as e:
            import traceback
            msg = f'{">" * 20} Trial {trial_item.trial_no} failed! {"<" * 20}\n' \
                  + f'{e.__class__.__name__}: {e}\n' \
                  + traceback.format_exc() \
                  + '*' * 50
            logger.error(msg)

            if self.on_trial_done:
                try:
                    self.on_trial_done(trial_item)
                except:
                    pass


class DaskDispatcher(Dispatcher):
    def __init__(self, work_dir):
        try:
            default_client()
        except ValueError:
            # create default Client
            # client = Client("tcp://127.0.0.1:55208")
            # client = Client(processes=False, threads_per_worker=5, n_workers=1, memory_limit='4GB')
            Client()  # detect env: DASK_SCHEDULER_ADDRESS

        super(DaskDispatcher, self).__init__()

        self.work_dir = work_dir
        self.models_dir = f'{work_dir}/models'

        fs.makedirs(self.models_dir, exist_ok=True)

    def dispatch(self, hyper_model, X, y, X_val, y_val, cv, num_folds, max_trials, dataset_id, trial_store,
                 **fit_kwargs):
        assert not any(dask.is_dask_collection(i) for i in (X, y, X_val, y_val)), \
            f'{self.__class__.__name__} does not support to run trial with dask collection.'

        queue_size = c.dask_search_queue
        worker_count = c.dask_search_executors
        retry_limit = c.trial_retry_limit

        failed_counter = Counter()
        success_counter = Counter()

        def on_trial_start(trial_item):
            trial_item.start_at = time.time()
            if logger.is_info_enabled():
                msg = f'Start trial {trial_item.trial_no}, space_id={trial_item.space_id}' \
                      + f',model_file={trial_item.model_file}'
                logger.info(msg)
            for callback in hyper_model.callbacks:
                # callback.on_build_estimator(hyper_model, space_sample, estimator, trial_no) #fixme
                callback.on_trial_begin(hyper_model, trial_item.space_sample, trial_item.trial_no)

        def on_trial_done(trial_item):
            trial_item.done_at = time.time()

            if trial_item.reward != 0 and not math.isnan(trial_item.reward):  # success
                improved = hyper_model.history.append(trial_item)
                for callback in hyper_model.callbacks:
                    callback.on_trial_end(hyper_model, trial_item.space_sample,
                                          trial_item.trial_no, trial_item.reward,
                                          improved, trial_item.elapsed)
                success_counter()
            else:
                for callback in hyper_model.callbacks:
                    callback.on_trial_error(hyper_model, trial_item.space_sample, trial_item.trial_no)
                failed_counter()

            if logger.is_info_enabled():
                elapsed = '%.3f' % (trial_item.done_at - trial_item.start_at)
                msg = f'Trial {trial_item.trial_no} done with reward={trial_item.reward}, ' \
                      f'elapsed {elapsed} seconds\n'
                logger.info(msg)
            if trial_store is not None:
                trial_store.put(dataset_id, trial_item)

        pool = DaskExecutorPool(worker_count, queue_size,
                                on_trial_start, on_trial_done,
                                hyper_model._run_trial,
                                X, y, X_val, y_val, fit_kwargs)
        pool.start()

        trial_no = 1
        retry_counter = 0

        while trial_no <= max_trials and pool.running:
            if pool.qsize >= queue_size:
                time.sleep(0.1)
                continue

            space_sample = hyper_model.searcher.sample()
            if hyper_model.history.is_existed(space_sample):
                if retry_counter >= retry_limit:
                    logger.info(f'Unable to take valid sample and exceed the retry limit 1000.')
                    break
                trial = hyper_model.history.get_trial(space_sample)
                for callback in hyper_model.callbacks:
                    callback.on_skip_trial(hyper_model, space_sample, trial_no, 'trial_existed', trial.reward, False,
                                           trial.elapsed)
                retry_counter += 1
                continue

            try:
                if trial_store is not None:
                    trial = trial_store.get(dataset_id, space_sample)
                    if trial is not None:
                        reward = trial.reward
                        elapsed = trial.elapsed
                        trial = Trial(space_sample, trial_no, reward, elapsed)
                        improved = hyper_model.history.append(trial)
                        hyper_model.searcher.update_result(space_sample, reward)
                        for callback in hyper_model.callbacks:
                            callback.on_skip_trial(hyper_model, space_sample, trial_no, 'hit_trial_store', reward,
                                                   improved,
                                                   elapsed)
                        trial_no += 1
                        continue

                model_file = '%s/%05d_%s.pkl' % (self.models_dir, trial_no, space_sample.space_id)

                item = DaskTrialItem(space_sample, trial_no, model_file=model_file)
                pool.push(item)

                if logger.is_info_enabled():
                    logger.info(f'Found trial {trial_no}, queue size: {pool.qsize}')
            except EarlyStoppingError:
                pool.stop()
                break
            except KeyboardInterrupt:
                pool.stop()
                pool.interrupted = True
                print('KeyboardInterrupt')
                break
            except Exception as e:
                import traceback
                msg = f'{">" * 20} Search trial {trial_no} failed! {"<" * 20}\n' \
                      + f'{e.__class__.__name__}: {e}\n' \
                      + traceback.format_exc() \
                      + '*' * 50
                logger.error(msg)
            finally:
                trial_no += 1
                retry_counter = 0

        # wait trials
        if pool.running:
            logger.info('Search done, wait trial tasks.')
        pool.push(None)  # mark end
        pool.join()

        if logger.is_info_enabled():
            logger.info(f'Search and all trials done, {success_counter.value} success, '
                        f'{failed_counter.value} failed.')

        return trial_no



# -*- coding:utf-8 -*-
__author__ = 'yangjian'
"""

"""

from hypernets.core.search_space import *
from hypernets.core.ops import *
import numpy as np
import pytest


class Test_Parameter:
    def test_alias(self):
        space = HyperSpace()
        with space.as_default():
            int1 = Int(1, 100)
            int2 = Int(1, 2)
            c1 = Choice([1, 2])

            assert int1.alias is None
            assert int2.alias is None
            assert c1.alias is None

            id1 = Identity(p1=int1, p2=int2, p3=c1)
            assert int1.alias == 'Module_Identity_1.p1'
            assert int2.alias == 'Module_Identity_1.p2'
            assert c1.alias == 'Module_Identity_1.p3'

            id2 = Identity(name='id2', action=int2)
            assert int2.alias == 'Module_Identity_1.p2,id2.action'

    def test_choice(self):
        state = np.random.RandomState(9527)
        c1 = Choice(['a', 'b', 'c', 'd'], random_state=state)

        v1 = c1.random_sample()
        assert v1 == c1.value
        assert v1 == 'a'

        c2 = Choice([1, 2, 3, 4, 5])
        c2.assign(1)
        assert c2.value == 1

        c3 = Choice([1, 'a', 'b', 4])
        with pytest.raises(AssertionError) as excinfo:
            c3.assign('f')

    def test_int(self):
        state = np.random.RandomState(9527)
        i1 = Int(0, 100, random_state=state)

        v1 = i1.random_sample()
        assert v1 == i1.value
        assert v1 == 97

        i2 = Int(0, 100)
        i2.assign(0)
        assert i2.value == 0

        i3 = Int(0, 100)
        i3.assign(100)
        assert i3.value == 100

        i4 = Int(0, 100)
        with pytest.raises(AssertionError) as excinfo:
            i4.assign(101)

    def test_real(self):
        state = np.random.RandomState(9527)
        r1 = Real(0, 1, prior="uniform", random_state=state)
        r2 = Real(0, 1, prior="q_uniform", q=0.2, random_state=state)
        r3 = Real(0.01, 1, prior="log_uniform", random_state=state)
        r4 = Real(0, 1)
        r5 = Real(0, 1)
        assert r1.assigned == False
        assert r4.assigned == False
        v1 = r1.random_sample()
        v2 = r2.random_sample()
        v3 = r3.random_sample()
        r4.assign(0.2)
        assert r1.assigned == True
        assert r4.assigned == True
        assert v1 == 0.33
        assert v2 == 0.4
        assert v3 == 1.27005017
        assert r4.value == 0.2

        with pytest.raises(AssertionError) as excinfo:
            r5.assign(2.0)

    def test_multiple_choice(self):
        mc1 = MultipleChoice(['a', 'b', 'c', 'd', 'e', 'f'], random_state=np.random.RandomState(1))
        mc2 = MultipleChoice(['a', 'b', 'c', 'd', 'e', 'f'], random_state=np.random.RandomState(2))
        mc3 = MultipleChoice([1, 2, 3, 4, 5], num_chosen_most=2, random_state=np.random.RandomState(3))
        mc4 = MultipleChoice(['a', 1, True, 'test', 5.6], num_chosen_most=1, random_state=np.random.RandomState(4))

        v1 = mc1.random_sample()
        v2 = mc2.random_sample()
        v3 = mc3.random_sample()
        v4 = mc4.random_sample()

        assert v1 == ['a', 'b', 'c', 'd', 'e', 'f']
        assert v2 == ['e']
        assert v3 == [4]
        assert v4 == ['a']

        with pytest.raises(AssertionError):
            MultipleChoice([1],num_chosen_least=2)

        mc5 = MultipleChoice([1, 2, 3, 4, 5], num_chosen_most=2, random_state=np.random.RandomState(3))
        with pytest.raises(AssertionError):
            mc5.assign([])

    def test_dynamic(self):
        i = Int(1, 100)
        r = Real(0, 10.0)
        c = Choice(['a', 'b', 'c'])

        d1 = Dynamic(lambda i1, r1, c1: f'i1:{i1}, r1:{r1}, c1:{c1}', i1=i, r1=r, c1=c)

        i.assign(33)
        assert d1.value is None
        assert d1.assigned == False

        r.assign(0.99)
        assert d1.value is None
        assert d1.assigned == False

        c.assign('b')
        assert d1.assigned == True
        assert d1.value == 'i1:33, r1:0.99, c1:b'

        d2 = Dynamic(
            lambda: 'no dependent'
        )

        assert d2.assigned == True
        assert d2.value == 'no dependent'

    def test_cascade(self):
        def get_space():
            space = HyperSpace()
            with space.as_default():
                c1 = Choice(['a', 'b', 'c'])

                def cc1_fn(c1, s):
                    with s.as_default():
                        if c1 == 'a':
                            return 'm1', Choice([1, 2])
                        elif c1 == 'b':
                            return 'm1', Choice([3, 4])
                        else:
                            return 'm1', Choice([5, 6])

                cc1 = Cascade(lambda args, space: cc1_fn(args['c1'], space), c1=c1)

                def cc2_fn(m1, s):
                    with s.as_default():
                        if isinstance(m1, ParameterSpace):
                            m1 = m1.value
                        if m1 == 5:
                            return 'm2', Choice([11, 22])
                        elif m1 == 6:
                            return 'm2', Choice([33, 44])
                        else:
                            return 'm2', Constant(-1)

                cc2 = Cascade(lambda args, space: cc2_fn(args['m1'], space), m1=cc1)

                def cc3_fn(m2, s):
                    with s.as_default():
                        if isinstance(m2, ParameterSpace):
                            m2 = m2.value
                        if m2 == 11:
                            return 'm3', Choice([55, 66])
                        else:
                            return 'm3', Constant(-1)

                cc3 = Cascade(lambda args, space: cc3_fn(args['m2'], space), m2=cc2)
                id1 = Identity(p1=c1, p2=cc1, p3=cc2, p4=cc3)
            return space

        space = get_space()
        space.Param_Choice_1.assign('a')
        assert space.Param_Cascade_1.value.options == [1, 2]

        space = get_space()
        space.Param_Choice_1.assign('b')
        assert space.Param_Cascade_1.value.options == [3, 4]

        space = get_space()
        space.Param_Choice_1.assign('c')
        assert space.Param_Cascade_1.value.options == [5, 6]

        assert space.Module_Identity_1.all_assigned == False
        space.Param_Choice_2.assign(5)

        assert space.Param_Cascade_2.value.options == [11, 22]

        assert space.Module_Identity_1.all_assigned == False
        space.Param_Choice_3.assign(22)

        assert space.Param_Cascade_3.value.value == -1
        assert space.Module_Identity_1.all_assigned == True

        assert len(space.get_assigned_params()) == 3

    def test_same(self):
        int1 = Int(1, 10)
        int2 = Int(1, 10)
        assert int1.same_config(int2)
        int2.alias = 'int2'
        assert not int1.same_config(int2)

        int3 = Int(0, 1)
        assert not int1.same_config(int3)

        real1 = Real(0, 1)
        real2 = Real(0, 1)

        assert real1.same_config(real2)

        real3 = Real(0, 1, q=1)
        assert not real1.same_config(real3)

        choice1 = Choice([1, 2])
        choice2 = Choice([1, 2])
        assert choice1.same_config(choice2)

        choice3 = Choice([1, 2, 3])
        assert not choice1.same_config(choice3)

        bool1 = Bool()
        bool2 = Bool()
        assert bool1.same_config(bool2)

        const1 = Constant(1)
        const2 = Constant(1)
        assert const1.same_config(const2)
        const3 = Constant('b')
        assert not const1.same_config(const3)

    def test_numeric2value(self):
        int1 = Int(1, 100)
        assert int1.value2numeric(10) == 10
        assert int1.numeric2value(10) == 10

        real1 = Real(0., 100.)
        assert real1.value2numeric(10.) == 10.
        assert real1.numeric2value(10.) == 10.

        bool1 = Bool()
        assert bool1.value2numeric(True) == 1
        assert bool1.numeric2value(1) == True

        choice1 = Choice(['a', 'b', 'c', 'd'])
        assert choice1.value2numeric('c') == 2
        assert choice1.numeric2value(2) == 'c'

        mutiple_choice1 = MultipleChoice(['a', 'b', 'c', 'd'])
        assert mutiple_choice1.value2numeric(['b', 'd']) == 5
        assert mutiple_choice1.numeric2value(5) == ['b', 'd']

    def test_label(self):
        space = HyperSpace()
        with space.as_default():
            int1 = Int(1, 100)
            assert int1.label == 'Param_Int_1-1-100-1'
            real1 = Real(0., 100.)
            assert real1.label == 'Param_Real_1-0.0-100.0-None-uniform-0.01'
            bool1 = Bool()
            assert bool1.label == 'Param_Bool_1-[False, True]'

            choice1 = Choice(['a', 'b', 'c', 'd'])
            assert choice1.label == 'Param_Choice_1-[\'a\', \'b\', \'c\', \'d\']'
            mutiple_choice1 = MultipleChoice(['a', 'b', 'c', 'd'])
            assert mutiple_choice1.label == 'Param_MultipleChoice_1-[\'a\', \'b\', \'c\', \'d\']-0-1'

    def test_expansion(self):
        int1 = Int(1, 100)
        vs = int1.expansion(20)
        assert len(vs) == 20
        vs = int1.expansion(101)
        assert len(vs) == 99
        vs = int1.expansion(0)
        assert len(vs) == 99

        real1 = Real(0., 100., max_expansion=10)
        vs = real1.expansion(20)
        assert len(vs) == 20
        vs = real1.expansion(0)
        assert len(vs) == 10

        bool1 = Bool()
        assert len(bool1.expansion()) == 2

        choice1 = Choice(['a', 'b', 'c', 'd'])
        assert len(choice1.expansion()) == 4

        mutiple_choice1 = MultipleChoice(['a', 'b', 'c', 'd'])
        vs = mutiple_choice1.expansion(20)
        assert len(vs) == 15

        vs = mutiple_choice1.expansion(0)
        assert len(vs) == 15

        mutiple_choice2 = MultipleChoice(['a', 'b', 'c', 'd'], num_chosen_most=1)
        vs = mutiple_choice2.expansion(4)
        assert len(vs) == 4
        vs = mutiple_choice2.expansion(20)
        assert len(vs) == 4



import json

from hypernets.experiment import CompeteExperiment
from hypernets.experiment import Experiment
from hypernets.tabular import dask_ex as dex
from hypernets.tests.model.plain_model_test import create_plain_model
from hypernets.tests.tabular.dask_transofromer_test import setup_dask

from hypernets.tabular.datasets import dsutils

import numpy as np
import pandas as pd

from dask import dataframe as dd
from dask.distributed import LocalCluster, Client
from sklearn.preprocessing import LabelEncoder


class Test_Get_character:
	@classmethod
	def setup_class(cls):
		setup_dask(cls)

		cls.boston = dd.from_pandas(dsutils.load_boston(), npartitions=1)
		cls.blood = dd.from_pandas(dsutils.load_blood(), npartitions=1)
		cls.bike_sharing = dd.from_pandas(dsutils.load_Bike_Sharing(), npartitions=1)

	# A test for multiclass task
	def experiment_with_bike_sharing(self, init_kwargs, run_kwargs, row_count=3000, with_dask=False):
		if with_dask:
			X = self.bike_sharing.copy()
			y = X.pop('count')
			y = y.astype('str')
		else:
			X = dsutils.load_Bike_Sharing()
			if row_count is not None:
				X = X.head(row_count)
			X['count'] = LabelEncoder().fit_transform(X['count'])
			y = X.pop('count')

		hyper_model = create_plain_model(with_encoder=True)

		X_train, X_test, y_train, y_test = \
			dex.train_test_split(X, y, test_size=0.3, random_state=9527)
		X_train, X_eval, y_train, y_eval = \
			dex.train_test_split(X_train, y_train, test_size=0.3, random_state=9527)

		init_kwargs = {
			'X_eval': X_eval, 'y_eval': y_eval, 'X_test': X_test,
			**init_kwargs
		}
		
		compete_experiment = CompeteExperiment(hyper_model, X_train, y_train, **init_kwargs)
		base_experiment = Experiment(hyper_model, X_train, y_train, **init_kwargs)

		mydict_compete = compete_experiment.get_data_character()
		mydict_base = base_experiment.get_data_character()

		assert mydict_base
		assert mydict_compete
		assert mydict_base['experimentType'] is 'base'
		assert mydict_compete['experimentType'] is 'compete'
		assert mydict_base['target']['taskType'] is 'multiclass'
		assert mydict_base['target']['freq'] is None
		assert mydict_base['target']['unique']
		assert mydict_base['target']['mean'] is None
		assert mydict_base['target']['max'] is None
		assert mydict_base['target']['min'] is None
		assert mydict_base['target']['stdev'] is None
		assert mydict_base['target']['dataType']
		assert len(mydict_base['targetDistribution']) <= 10
		assert mydict_base['datasetShape']['X_train']
		assert mydict_base['datasetShape']['y_train']
		assert mydict_base['datasetShape']['X_eval']
		assert mydict_base['datasetShape']['y_eval']
		assert mydict_base['datasetShape']['X_test']
		assert mydict_compete['featureDistribution']
	
	# A test for binary task
	def experiment_with_blood(self, init_kwargs, run_kwargs, row_count=3000, with_dask=False):
		if with_dask:
			X = self.blood.copy()
			y = X.pop('Class')
		else:
			X = dsutils.load_blood()
			if row_count is not None:
				X = X.head(row_count)
			X['Class'] = LabelEncoder().fit_transform(X['Class'])
			y = X.pop('Class')
		
		hyper_model = create_plain_model(with_encoder=True)

		X_train, X_test, y_train, y_test = \
		dex.train_test_split(X, y, test_size=0.3, random_state=9527)
		X_train, X_eval, y_train, y_eval = \
		dex.train_test_split(X_train, y_train, test_size=0.3, random_state=9527)

		init_kwargs = {
			'X_eval': X_eval, 'y_eval': y_eval, 'X_test': X_test,
			**init_kwargs
		}

		compete_experiment = CompeteExperiment(hyper_model, X_train, y_train, **init_kwargs)
		base_experiment = Experiment(hyper_model, X_train, y_train, **init_kwargs)

		mydict_compete = compete_experiment.get_data_character()
		mydict_base = base_experiment.get_data_character()

		assert mydict_base
		assert mydict_compete
		assert mydict_base['experimentType'] is 'base'
		assert mydict_compete['experimentType'] is 'compete'
		assert mydict_base['target']['taskType'] is 'binary'
		assert mydict_base['target']['freq'] is not None
		assert mydict_base['target']['unique'] is 2
		assert mydict_base['target']['mean'] is None
		assert mydict_base['target']['max'] is None
		assert mydict_base['target']['min'] is None
		assert mydict_base['target']['stdev'] is None
		assert mydict_base['target']['dataType']
		assert len(mydict_base['targetDistribution']) <= 10
		assert mydict_base['datasetShape']['X_train']
		assert mydict_base['datasetShape']['y_train']
		assert mydict_base['datasetShape']['X_eval']
		assert mydict_base['datasetShape']['y_eval']
		assert mydict_base['datasetShape']['X_test']
		assert mydict_compete['featureDistribution']

	# A test for regression task
	def experiment_with_boston(self, init_kwargs, run_kwargs, row_count=3000, with_dask=False):
		if with_dask:
			X = self.boston
			y = X.pop('target')
		else:
			X = dsutils.load_boston()
			if row_count is not None:
				X = X.head(row_count)
			X['target'] = LabelEncoder().fit_transform(X['target'])
			y = X.pop('target')
			y = y.astype('float64')
		
		hyper_model = create_plain_model(with_encoder=True)
		
		X_train, X_test, y_train, y_test = \
		dex.train_test_split(X, y, test_size=0.3, random_state=9527)
		X_train, X_eval, y_train, y_eval = \
		dex.train_test_split(X_train, y_train, test_size=0.3, random_state=9527)

		init_kwargs = {
			'X_eval': X_eval, 'y_eval': y_eval, 'X_test': X_test,
			**init_kwargs
		}

		compete_experiment = CompeteExperiment(hyper_model, X_train, y_train, **init_kwargs)
		base_experiment = Experiment(hyper_model, X_train, y_train, **init_kwargs)

		mydict_compete = compete_experiment.get_data_character()
		mydict_base = base_experiment.get_data_character()

		assert mydict_base
		assert mydict_compete
		assert mydict_base['experimentType'] is 'base'
		assert mydict_compete['experimentType'] is 'compete'
		assert mydict_base['target']['taskType'] is 'regression'
		assert mydict_base['target']['freq'] is None
		assert mydict_base['target']['unique']
		assert mydict_base['target']['mean'] is not None
		assert mydict_base['target']['max'] is not None
		assert mydict_base['target']['min'] is not None
		assert mydict_base['target']['stdev'] is not None
		assert mydict_base['target']['dataType'] is 'float'
		assert len(mydict_base['targetDistribution']) <= 10
		assert mydict_base['datasetShape']['X_train']
		assert mydict_base['datasetShape']['y_train']
		assert mydict_base['datasetShape']['X_eval']
		assert mydict_base['datasetShape']['y_eval']
		assert mydict_base['datasetShape']['X_test']
		assert mydict_compete['featureDistribution']

	def test_multiclass_with_bike_sharing(self):
		self.experiment_with_bike_sharing({}, {})
		self.experiment_with_bike_sharing({}, {}, with_dask = True)
	
	def test_binary_with_blood(self):
		self.experiment_with_blood({}, {})
		self.experiment_with_blood({}, {}, with_dask = True)

	def test_regression_with_boston(self):

		self.experiment_with_boston({}, {})
		self.experiment_with_boston({}, {}, with_dask = True)



# -*- coding:utf-8 -*-
__author__ = 'yangjian'
"""

"""
from collections import OrderedDict

import pandas as pd
from IPython.display import display, display_markdown

from hypernets.tabular import dask_ex as dex
from hypernets.utils import const
from . import ExperimentCallback
from .compete import StepNames


class ConsoleCallback(ExperimentCallback):
    def experiment_start(self, exp):
        print('experiment start')

    def experiment_end(self, exp, elapsed):
        print(f'experiment end')
        print(f'   elapsed:{elapsed}')

    def experiment_break(self, exp, error):
        print(f'experiment break, error:{error}')

    def step_start(self, exp, step):
        print(f'   step start, step:{step}')

    def step_progress(self, exp, step, progress, elapsed, eta=None):
        print(f'      progress:{progress}')
        print(f'         elapsed:{elapsed}')

    def step_end(self, exp, step, output, elapsed):
        print(f'   step end, step:{step}, output:{output.items() if output is not None else ""}')
        print(f'      elapsed:{elapsed}')

    def step_break(self, exp, step, error):
        print(f'step break, step:{step}, error:{error}')


class StepCallback:
    def step_start(self, exp, step):
        pass

    def step_progress(self, exp, step, progress, elapsed, eta=None):
        pass

    def step_end(self, exp, step, output, elapsed):
        pass

    def step_break(self, exp, step, error):
        pass


class NotebookStepCallback(StepCallback):
    def step_start(self, exp, step):
        title = step.replace('_', ' ').title()
        display_markdown(f'## {title}', raw=True)

        params = exp.get_step(step).get_params()
        data = self.params_to_display(params)
        if data is not None:
            display_markdown('### Initliazed parameters', raw=True)
            display(data, display_id=f'input_{step}')

    def step_end(self, exp, step, output, elapsed):
        step_obj = exp.get_step(step)
        fitted_params = step_obj.get_fitted_params()
        data = self.fitted_params_to_display(fitted_params)
        if data is not None:
            display_markdown('### Fitted parameters', raw=True)
            display(data, display_id=f'output_{step}')

        display_markdown('### Elapsed', raw=True)
        display_markdown(f'* {step_obj.elapsed_seconds:.3f} seconds', raw=True)

    @staticmethod
    def params_to_display(params):
        df = pd.DataFrame(params.items(), columns=['key', 'value'])
        return df

    @staticmethod
    def fitted_params_to_display(fitted_params):
        df = pd.DataFrame(fitted_params.items(), columns=['key', 'value'])
        return df


class NotebookFeatureSelectionStepCallback(NotebookStepCallback):
    @staticmethod
    def fitted_params_to_display(fitted_params):
        params = fitted_params.copy()

        selected = params.get('selected_features')
        unselected = params.get('unselected_features')
        if selected is not None and unselected is not None:
            params['kept/dropped feature count'] = f'{len(selected)}/{len(unselected)}'

        return NotebookStepCallback.fitted_params_to_display(params)


class NotebookFeatureImportanceSelectionStepCallback(NotebookStepCallback):

    def step_end(self, exp, step, output, elapsed):
        super().step_end(exp, step, output, elapsed)

        fitted_params = exp.get_step(step).get_fitted_params()
        input_features = fitted_params.get('input_features')
        selected = fitted_params.get('selected_features')
        if selected is None:
            selected = input_features
        importances = fitted_params.get('importances', [])
        is_selected = [input_features[i] in selected for i in range(len(importances))]
        df = pd.DataFrame(
            zip(input_features, importances, is_selected),
            columns=['feature', 'importance', 'selected'])
        df = df.sort_values('importance', axis=0, ascending=False)

        display_markdown('### Feature importances', raw=True)
        display(df, display_id=f'output_{step}_importances')


class NotebookPermutationImportanceSelectionStepCallback(NotebookStepCallback):

    def step_end(self, exp, step, output, elapsed):
        super().step_end(exp, step, output, elapsed)

        fitted_params = exp.get_step(step).get_fitted_params()
        importances = fitted_params.get('importances')
        if importances is not None:
            df = pd.DataFrame(
                zip(importances['columns'], importances['importances_mean'], importances['importances_std']),
                columns=['feature', 'importance', 'std'])

            display_markdown('### Permutation importances', raw=True)
            display(df, display_id=f'output_{step}_importances')


class NotebookEstimatorBuilderStepCallback(NotebookStepCallback):
    @staticmethod
    def fitted_params_to_display(fitted_params):
        return fitted_params.get('estimator')


class NotebookPseudoLabelingStepCallback(NotebookStepCallback):
    def step_end(self, exp, step, output, elapsed):
        super().step_end(exp, step, output, elapsed)

        try:
            fitted_params = exp.get_step(step).get_fitted_params()
            proba = fitted_params.get('test_proba')
            proba = proba[:, 1]  # fixme for multi-classes

            import seaborn as sns
            import matplotlib.pyplot as plt
            # Draw Plot
            plt.figure(figsize=(8, 4), dpi=80)
            sns.kdeplot(proba, shade=True, color="g", label="Proba", alpha=.7, bw_adjust=0.01)
            # Decoration
            plt.title('Density Plot of Probability', fontsize=22)
            plt.legend()
            plt.show()
        except:
            pass


_default_step_callback = NotebookStepCallback

_step_callbacks = {
    StepNames.DATA_CLEAN: NotebookFeatureSelectionStepCallback,
    StepNames.FEATURE_GENERATION: NotebookStepCallback,
    StepNames.MULITICOLLINEARITY_DETECTION: NotebookFeatureSelectionStepCallback,
    StepNames.DRIFT_DETECTION: NotebookFeatureSelectionStepCallback,
    StepNames.FEATURE_IMPORTANCE_SELECTION: NotebookFeatureImportanceSelectionStepCallback,
    StepNames.SPACE_SEARCHING: NotebookStepCallback,
    StepNames.ENSEMBLE: NotebookEstimatorBuilderStepCallback,
    StepNames.TRAINING: NotebookEstimatorBuilderStepCallback,
    StepNames.PSEUDO_LABELING: NotebookPseudoLabelingStepCallback,
    StepNames.FEATURE_RESELECTION: NotebookPermutationImportanceSelectionStepCallback,
    StepNames.FINAL_SEARCHING: NotebookStepCallback,
    StepNames.FINAL_ENSEMBLE: NotebookEstimatorBuilderStepCallback,
    StepNames.FINAL_TRAINING: NotebookEstimatorBuilderStepCallback,
}


class SimpleNotebookCallback(ExperimentCallback):
    def __init__(self):
        super(SimpleNotebookCallback, self).__init__()

        self.exp = None
        self.steps = None
        self.running = None

    def experiment_start(self, exp):
        self.exp = exp
        self.steps = OrderedDict()
        self.running = True

        display_markdown('### Input Data', raw=True)

        X_train, y_train, X_test, X_eval, y_eval = \
            exp.X_train, exp.y_train, exp.X_test, exp.X_eval, exp.y_eval

        if dex.exist_dask_object(X_train, y_train, X_test, X_eval, y_eval):
            display_data = (dex.compute(X_train.shape)[0],
                            dex.compute(y_train.shape)[0],
                            dex.compute(X_eval.shape)[0] if X_eval is not None else None,
                            dex.compute(y_eval.shape)[0] if y_eval is not None else None,
                            dex.compute(X_test.shape)[0] if X_test is not None else None,
                            exp.task if exp.task == const.TASK_REGRESSION
                            else f'{exp.task}({dex.compute(y_train.nunique())[0]})')
        else:
            display_data = (X_train.shape,
                            y_train.shape,
                            X_eval.shape if X_eval is not None else None,
                            y_eval.shape if y_eval is not None else None,
                            X_test.shape if X_test is not None else None,
                            exp.task if exp.task == const.TASK_REGRESSION
                            else f'{exp.task}({y_train.nunique()})')
        display(pd.DataFrame([display_data],
                             columns=['X_train.shape',
                                      'y_train.shape',
                                      'X_eval.shape',
                                      'y_eval.shape',
                                      'X_test.shape',
                                      'Task', ]), display_id='output_intput')

        try:
            import seaborn as sns
            import matplotlib.pyplot as plt
            from sklearn.preprocessing import LabelEncoder

            le = LabelEncoder()
            y = le.fit_transform(y_train.dropna())
            # Draw Plot
            plt.figure(figsize=(8, 4), dpi=80)
            sns.distplot(y, kde=False, color="g", label="y")
            # Decoration
            plt.title('Distribution of y', fontsize=12)
            plt.legend()
            plt.show()
        except:
            pass

    def experiment_end(self, exp, elapsed):
        self.running = False

    def experiment_break(self, exp, error):
        self.running = False

    def step_start(self, exp, step):
        cb_cls = _step_callbacks[step] if step in _step_callbacks.keys() else _default_step_callback
        cb = cb_cls()
        cb.step_start(exp, step)
        self.steps[step] = cb

    def step_progress(self, exp, step, progress, elapsed, eta=None):
        assert self.steps is not None and step in self.steps.keys()
        cb = self.steps[step]
        cb.step_progress(exp, step, progress, elapsed, eta=eta)

    def step_end(self, exp, step, output, elapsed):
        assert self.steps is not None and step in self.steps.keys()
        cb = self.steps[step]
        cb.step_end(exp, step, output, elapsed)

    def step_break(self, exp, step, error):
        assert self.steps is not None and step in self.steps.keys()
        cb = self.steps[step]
        cb.step_break(exp, step, error)



from datetime import datetime

from sklearn.metrics import get_scorer
from sklearn.preprocessing import LabelEncoder

from hypernets.experiment import CompeteExperiment
from hypernets.tabular import dask_ex as dex
from hypernets.tabular.datasets import dsutils
from hypernets.tabular.metrics import calc_score, metric_to_scoring
from hypernets.tests.model.plain_model_test import create_plain_model
from hypernets.tests.tabular.dask_transofromer_test import setup_dask


def experiment_with_bank_data(init_kwargs, run_kwargs, row_count=3000, with_dask=False):
    hyper_model = create_plain_model(with_encoder=True, with_dask=with_dask)
    scorer = get_scorer(metric_to_scoring(hyper_model.reward_metric))
    X = dsutils.load_bank()
    if row_count is not None:
        X = X.head(row_count)
    X['y'] = LabelEncoder().fit_transform(X['y'])

    if with_dask:
        setup_dask(None)
        X = dex.dd.from_pandas(X, npartitions=1)

    y = X.pop('y')

    X_train, X_test, y_train, y_test = \
        dex.train_test_split(X, y, test_size=0.3, random_state=9527)
    X_train, X_eval, y_train, y_eval = \
        dex.train_test_split(X_train, y_train, test_size=0.3, random_state=9527)

    init_kwargs = {
        'X_eval': X_eval, 'y_eval': y_eval, 'X_test': X_test,
        'scorer': scorer,
        'ensemble_size': 0,
        'drift_detection': False,
        **init_kwargs
    }
    run_kwargs = {
        'max_trials': 3,
        **run_kwargs
    }
    experiment = CompeteExperiment(hyper_model, X_train, y_train, **init_kwargs)
    estimator = experiment.run(**run_kwargs)

    assert estimator

    preds = estimator.predict(X_test)
    proba = estimator.predict_proba(X_test)

    if with_dask:
        preds, proba = dex.compute(preds, proba)

    score = calc_score(y_test, preds, proba, metrics=['auc', 'accuracy', 'f1', 'recall', 'precision'])
    print('evaluate score:', score)
    assert score


def experiment_with_movie_lens(init_kwargs, run_kwargs, row_count=None, with_dask=False):
    hyper_model = create_plain_model(reward_metric='f1', with_encoder=True, with_dask=with_dask)
    scorer = get_scorer(metric_to_scoring(hyper_model.reward_metric))
    X = dsutils.load_movielens()
    # X['genres'] = X['genres'].apply(lambda s: s.replace('|', ' '))
    X['timestamp'] = X['timestamp'].apply(datetime.fromtimestamp)
    if row_count is not None:
        X = X.head(row_count)

    if with_dask:
        setup_dask(None)
        X = dex.dd.from_pandas(X, npartitions=1)

    y = X.pop('rating')

    X_train, X_test, y_train, y_test = \
        dex.train_test_split(X, y, test_size=0.3, random_state=9527)
    X_train, X_eval, y_train, y_eval = \
        dex.train_test_split(X_train, y_train, test_size=0.3, random_state=9527)

    init_kwargs = {
        'X_eval': X_eval, 'y_eval': y_eval, 'X_test': X_test,
        'scorer': scorer,
        'ensemble_size': 0,
        'drift_detection': False,
        **init_kwargs
    }
    run_kwargs = {
        'max_trials': 3,
        **run_kwargs
    }
    experiment = CompeteExperiment(hyper_model, X_train, y_train, **init_kwargs)
    estimator = experiment.run(**run_kwargs)

    assert estimator

    preds = estimator.predict(X_test)
    proba = estimator.predict_proba(X_test)

    if with_dask:
        preds, proba = dex.compute(preds, proba)

    score = calc_score(y_test, preds, proba,
                       metrics=['auc', 'accuracy', 'f1', 'recall', 'precision'],
                       task=experiment.task)
    print('evaluate score:', score)
    assert score


def test_simple():
    experiment_with_bank_data({}, {})


def test_without_eval():
    experiment_with_bank_data(dict(X_eval=None, y_eval=None, ), {})


def test_without_xtest():
    experiment_with_bank_data(dict(X_test=None), {})


def test_without_eval_xtest():
    experiment_with_bank_data(dict(X_eval=None, y_eval=None, X_test=None), {})


def test_with_adversarial_validation():
    experiment_with_bank_data(dict(cv=False,
                                   X_eval=None,
                                   y_eval=None,
                                   train_test_split_strategy='adversarial_validation'), {})


def test_with_ensemble():
    experiment_with_bank_data(dict(ensemble_size=5), {})


def test_without_cv():
    experiment_with_bank_data(dict(cv=False), {})


def test_with_feature_generation():
    experiment_with_movie_lens(dict(feature_generation=True,
                                    feature_generation_text_cols=['title']), {})


def test_with_dd():
    experiment_with_bank_data(dict(drift_detection=True), {})


def test_with_cd():
    experiment_with_bank_data(dict(collinearity_detection=True), {})


def test_with_fi_threshold():
    experiment_with_bank_data(dict(feature_selection=True,
                                   feature_selection_threshold=0.0000001), {})


def test_with_fi_quantile():
    experiment_with_bank_data(dict(feature_selection=True,
                                   feature_selection_strategy='quantile',
                                   feature_selection_quantile=0.4), {})


def test_with_fi_number():
    experiment_with_bank_data(dict(feature_selection=True,
                                   feature_selection_strategy='number',
                                   feature_selection_number=10), {})


def test_with_pl_threshold():
    experiment_with_bank_data(dict(pseudo_labeling=True), {})


def test_with_pl_quantile():
    experiment_with_bank_data(dict(pseudo_labeling=True,
                                   pseudo_labeling_strategy='quantile'), {})


def test_with_pl_number():
    experiment_with_bank_data(dict(drift_detection=False, pseudo_labeling=True,
                                   pseudo_labeling_strategy='number'), {})


def test_with_pi():
    experiment_with_bank_data(dict(feature_reselection=True,
                                   feature_reselection_threshold=0.0001), {})


def test_with_pl_dask():
    experiment_with_bank_data(dict(cv=False, pseudo_labeling=True), {},
                              with_dask=True)


def test_with_ensemble_dask():
    experiment_with_bank_data(dict(ensemble_size=5, cv=False), {},
                              with_dask=True)


def test_with_cv_ensemble_dask():
    experiment_with_bank_data(dict(ensemble_size=5, cv=True), {},
                              row_count=6000, with_dask=True)



# -*- coding:utf-8 -*-
"""

"""
__author__ = 'yangjian'

from .evolution_searcher import EvolutionSearcher
from .mcts_searcher import MCTSSearcher
from .random_searcher import RandomSearcher
from .playback_searcher import PlaybackSearcher
from .grid_searcher import GridSearcher
from ..core.searcher import Searcher

searcher_dict = {
    'mcts': MCTSSearcher,
    'MCTS': MCTSSearcher,
    'MCTSSearcher': MCTSSearcher,
    'evolution': EvolutionSearcher,
    'Evolution': EvolutionSearcher,
    'EvolutionSearcher': EvolutionSearcher,
    'Random': RandomSearcher,
    'RandomSearcher': RandomSearcher,
    'random': RandomSearcher,
    'grid': GridSearcher,
    'Grid': GridSearcher,
    'GridSearcher': GridSearcher,
    'playback': PlaybackSearcher,
    'PlaybackSearcher': PlaybackSearcher,
    'Playback': PlaybackSearcher
}


def get_searcher_cls(identifier):
    if isinstance(identifier, str):
        cls = searcher_dict.get(identifier, None)
        if cls is None:
            raise ValueError(f'Illegal identifier:{identifier}')
        else:
            return cls
    elif isinstance(identifier, type):
        if issubclass(identifier, Searcher):
            return identifier
        else:
            raise ValueError(f'Wrong searcher type:{identifier}')
    else:
        raise ValueError(f'Illegal identifier:{identifier}')


def make_searcher(cls, search_space_fn, optimize_direction='min', **kwargs):
    cls = get_searcher_cls(cls)

    if cls == EvolutionSearcher:
        default_kwargs = dict(population_size=30, sample_size=10, candidates_size=10,
                              regularized=True, use_meta_learner=True)
    elif cls == MCTSSearcher:
        default_kwargs = dict(max_node_space=10)
    else:
        default_kwargs = {}
    kwargs = {**default_kwargs, **kwargs}
    searcher = cls(search_space_fn, optimize_direction=optimize_direction, **kwargs)
    return searcher
